# B isotope fns

import numpy as np
from cbsyst.helpers import NnotNone, Bunch
from .boron import chiB_calc

def get_alphaB():
    """
    Klochko alpha for B fractionation
    """
    return 1.0272

def get_epsilonB():
    """
    Klochko epsilon for B fractionation
    """
    return alpha_to_epsilon(get_alphaB())

def alpha_to_epsilon(alphaB):
    """
    Convert alpha to epsilon (which is alpha in delta space)

    Parameters
    ----------
    alphaB : array-like
        The isotope fractionation factor for (11/10 BO3)/(11/10 BO4).

    Returns
    -------
    array-like
        alphaB expressed in delta notation (AKA epsilonB).
    """
    return (alphaB-1)*1000

def epsilon_to_alpha(epsilonB):
    """
    Convert epsilon to alpha

    Parameters
    ----------
    epsilonB : array-like
        The isotope fractionation factor (11/10 BO3)/(11/10 BO4) expressed in delta notation.

    Returns
    -------
    array-like
        The isotope fractionation factor (11/10 BO3)/(11/10 BO4).
    """
    return (epsilonB/1000)+1


# Isotope Unit Converters
def A11_to_d11(A11, SRM_ratio=4.04367):
    """
    Convert fractional abundance (A11) to delta notation (d11).

    Parameters
    ----------
    A11 : array-like
        The fractional abundance of 11B: 11B / (11B + 10B).
    SRM_ratio : float, optional
        The 11B/10B of the SRM, by default NIST951 which is 4.04367

    Returns
    -------
    array-like
        A11 expressed in delta notation (d11).
    """
    return ((A11 / (1 - A11)) / SRM_ratio - 1) * 1000

def A11_to_R11(A11):
    """
    Convert fractional abundance (A11) to isotope ratio (R11).

    Parameters
    ----------
    A11 : array-like
        The fractional abundance of 11B: 11B / (11B + 10B).

    Returns
    -------
    array-like
        A11 expressed as an isotope ratio (R11).
    """
    return A11 / (1 - A11)

def d11_to_A11(d11, SRM_ratio=4.04367):
    """
    Convert delta notation (d11) to fractional abundance (A11).

    Parameters
    ----------
    d11 : array-like
        The isotope ratio expressed in delta notation.
    SRM_ratio : float, optional
        The 11B/10B of the SRM, by default NIST951 which is 4.04367

    Returns
    -------
    array-like
       Delta notation (d11) expressed as fractional abundance (A11).
    """
    return SRM_ratio * (d11 / 1000 + 1) / (SRM_ratio * (d11 / 1000 + 1) + 1)

def d11_to_R11(d11, SRM_ratio=4.04367):
    """
    Convert delta notation (d11) to isotope ratio (R11).

    Parameters
    ----------
    d11 : array-like
        The isotope ratio expressed in delta notation.
    SRM_ratio : float, optional
        The 11B/10B of the SRM, by default NIST951 which is 4.04367

    Returns
    -------
    array-like
       Delta notation (d11) expressed as isotope ratio (R11).
    """
    return (d11 / 1000 + 1) * SRM_ratio

def R11_to_d11(R11, SRM_ratio=4.04367):
    """
    Convert isotope ratio (R11) to delta notation (d11).

    Parameters
    ----------
    R11 : array-like
        The isotope ratio (11B/10B).
    SRM_ratio : float, optional
        The 11B/10B of the SRM, by default NIST951 which is 4.04367

    Returns
    -------
    array-like
        R11 expressed in delta notation (d11).
    """
    return (R11 / SRM_ratio - 1) * 1000

def R11_to_A11(R11):
    """
    Convert isotope ratio (R11) to fractional abundance (A11).

    Parameters
    ----------
    R11 : array-like
        The isotope ratio (11B/10B).

    Returns
    -------
    array-like
        R11 expressed as fractional abundance (A11).
    """
    return R11 / (1 + R11)

# Alpha Converters
def ABO3_to_ABO4(ABO3,alphaB):
    """
    Converts isotope fractional abundance of boric acid to isotope fraction abundance of borate ion

    Parameters
    ----------
    ABO3 : array-like
        The fractional abundance of boric acid (B(OH)3)
    alphaB : array-like
        The isotope fractionation factor for (11/10 BO3)/(11/10 BO4).

    Returns
    -------
    array-like
        ABO4 - the fractional abundance of borate ion (B(OH)4)
    """
    return (1 / ((alphaB / ABO3) - alphaB + 1) )

def ABO3_or_ABO4(ABO3,ABO4,alphaB):
    """
    Helper function to determine ABO4 if ABO3 is None

    Parameters
    ----------
    ABO3 : array-like
        The fractional abundance of boric acid (B(OH)3)
    ABO4 : array-like
        The fractional abundance of borate ion (B(OH)4)
    alphaB : array-like
        The isotope fractionation factor for (11/10 BO3)/(11/10 BO4).

    Returns
    -------
    array-like
        ABO4 - the fractional abundance of borate ion (B(OH)4)
    """
    if NnotNone(ABO3, ABO4) < 1:
        raise(ValueError("Either ABO4 or ABO3 must be specified"))
    elif ABO4 is None:
        ABO4 = ABO3_to_ABO4(ABO3,alphaB)
    return ABO4


# Base Functions
# Calculate total boron isotope fractional abundance using borate ion (B(OH)4)
def calculate_ABT(H, Ks, alphaB, ABO4=None, ABO3=None):
    """
    Calculate ABT from pH (total scale) and ABO4 or ABO3.

    Parameters
    ----------    
    Ks : dict
        A dictionary of stoichiometric equilibrium constants.
    pH : array-like
        pH on the Total scale
    alphaB : array-like
        The fractionation factor between B(OH)3 and B(OH)4-
    ABO4 : array-like
        The fractional abundance of 11B in B(OH)4.
    ABO3 : array-like
        The fractional abundance of 11B in B(OH)3.

    Returns
    -------
    array-like
        The fractional abundance of 11B in total B (ABT).
    """
    ABO4 = ABO3_or_ABO4(ABO3,ABO4,alphaB)

    chiB = chiB_calc(H, Ks)
    return (
        ABO4
        * (
            -ABO4 * alphaB * chiB
            + ABO4 * alphaB
            + ABO4 * chiB
            - ABO4
            + alphaB * chiB
            - chiB
            + 1
        )
        / (ABO4 * alphaB - ABO4 + 1))

# Calculate pH using isotope fractional abundance of borate ion (B(OH)4)
def calculate_H(Ks, alphaB, ABT, ABO4=None, ABO3=None):
    """
    Calculate H from ABO4 or ABO3 and ABT. 

    Parameters
    ----------
    Ks : dict
        dictionary of speciation constants
    alphaB : float or array-like
        fractionation factor between B(OH)3 and B(OH)4-
    ABT : float or array-like
        fractional abundance of 11B in total B
    ABO4 : float or array-like
        fractional abundance of 11B in B(OH)4-
        
    Returns
    -------
    array-like
        pH on the total scale.
    """
    ABO4 = ABO3_or_ABO4(ABO3,ABO4,alphaB)

    return (Ks.KB / ((alphaB / (1 - ABO4 + alphaB * ABO4) - 1) / (ABT / ABO4 - 1) - 1))

# Calculate isotope fractional abundance of boric acid (B(OH)3)
def calculate_ABO3(H, Ks, ABT, alphaB):
    """
    Calculate ABO3 from H and ABT

    Parameters
    ----------
    Ks : dict
        A dictionary of stoichiometric equilibrium constants.
    H : array-like
        The activity of Hydrogen ions in mol kg-1
    ABT : array-like
        The fractional abundance of 11B in total B.
    alphaB : array-like
        The fractionation factor between B(OH)3 and B(OH)4-

    Returns
    -------
    array-like
        The fractional abundance of 11B in B(OH)3.
    """
    chiB = chiB_calc(H, Ks)
    return (
        ABT * alphaB
        - ABT
        + alphaB * chiB
        - chiB
        - np.sqrt(
            ABT ** 2 * alphaB ** 2
            - 2 * ABT ** 2 * alphaB
            + ABT ** 2
            - 2 * ABT * alphaB ** 2 * chiB
            + 2 * ABT * alphaB
            + 2 * ABT * chiB
            - 2 * ABT
            + alphaB ** 2 * chiB ** 2
            - 2 * alphaB * chiB ** 2
            + 2 * alphaB * chiB
            + chiB ** 2
            - 2 * chiB
            + 1
        )
        + 1
    ) / (2 * chiB * (alphaB - 1))

# Calculate isotope fractional abundance of borate ion (B(OH)4)
def calculate_ABO4(H, Ks, ABT, alphaB):
    """
    Calculate ABO4 from H and ABT

    Parameters
    ----------
    Ks : dict
        Dictionary of stoichiometric equilibrium constants.
    H : array-like
        The activity of Hydrogen ions in mol kg-1
    ABT : array-like
        The fractional abundance of 11B in total B.
    alphaB : array-like
        The fractionation factor between B(OH)3 and B(OH)4-

    Returns
    -------
    array-like
        The fractional abundance of 11B in B(OH)4-
    """
    chiB = chiB_calc(H, Ks)
    return -(
        ABT * alphaB
        - ABT
        - alphaB * chiB
        + chiB
        + np.sqrt(
            ABT ** 2 * alphaB ** 2
            - 2 * ABT ** 2 * alphaB
            + ABT ** 2
            - 2 * ABT * alphaB ** 2 * chiB
            + 2 * ABT * alphaB
            + 2 * ABT * chiB
            - 2 * ABT
            + alphaB ** 2 * chiB ** 2
            - 2 * alphaB * chiB ** 2
            + 2 * alphaB * chiB
            + chiB ** 2
            - 2 * chiB
            + 1
        )
        - 1
    ) / (2 * alphaB * chiB - 2 * alphaB - 2 * chiB + 2)

# Calculate alpha using isotope fractional abundance of boric acid (B(OH)3)
def calculate_alpha_ABO3(H, Ks, ABT, ABO3):
    """
    Calculate fractionation factor (alpha) from the fractional abundance of 11B in B(OH)3 (ABO3)

    Parameters
    ----------
    Ks : dict
        Dictionary of stoichiometric equilibrium constants.
    H : array-like
        The activity of Hydrogen ions in mol kg-1
    ABT : array-like
        The fractional abundance of 11B in total B.
    ABO3 : array-like
        The fractional abundance of 11B in boric acid (B(OH)3).

    Returns
    -------
    array-like
        The fractionation factor between B(OH)3 and B(OH)4- (alpha)
    """
    return ( (1
            / ((H/Ks.KB) * (ABT - ABO3) + ABT)) 
            / (ABO3 -1))

# Calculate alpha using isotope fractional abundance of borate ion (B(OH)4)
def calculate_alpha_ABO4(H, Ks, ABT, ABO4):
    """
    Calculate fractionation factor (alpha) from the fractional abundance of 11B in B(OH)3 (ABO3)

    Parameters
    ----------
    Ks : dict
        Dictionary of stoichiometric equilibrium constants.
    H : array-like
        The activity of Hydrogen ions in mol kg-1
    ABT : array-like
        The fractional abundance of 11B in total B.
    ABO4 : array-like
        The fractional abundance of 11B in borate ion (B(OH)4).

    Returns
    -------
    array-like
        The fractionation factor between B(OH)3 and B(OH)4- (alpha)
    """
    return ( (1/ABO4 - 1)
            / (1 / (ABT - ((ABO4-ABT)/(H/Ks.KB))) -1) )

# Calculate alpha using isotope fractional abundance of borate ion (B(OH)4)
def calculate_KB(H, alphaB, ABT, ABO4=None, ABO3=None):
    """
    Calculate stoichiometric equilibrium constant for boron

    Parameters
    ----------
    H : array-like
        The activity of Hydrogen ions in mol kg-1
    alphaB : array-like
        The fractionation factor between B(OH)3 and B(OH)4-
    ABT : array-like
        The fractional abundance of 11B in total B.
    ABO4 : array-like
        The fractional abundance of 11B in borate ion (B(OH)4).
    ABO3 : array-like
        The fractional abundance of 11B in boric acid (B(OH)3).

    Returns
    -------
    array-like
        The stoichiometric equilibrium constant for boron (KB)
    """
    ABO4 = ABO3_or_ABO4(ABO3,ABO4,alphaB)
    return (H
            / ((ABO4 - ABT)
            / ( ABT 
            - 1 / ( (1/alphaB) * (1/ABO4 -1) + 1) )))

def calc_B_isotopes(pHtot=None, ABT=None, ABO3=None, ABO4=None, alphaB=None, Ks=None, **kwargs):
    # determine pH and ABT
    if pHtot is not None:  # pH is known
        H = 10.0**-pHtot
        if ABT is None:
            ABT = calculate_ABT(H=H, Ks=Ks, alphaB=alphaB, ABO3=ABO3, ABO4=ABO4)
    else:  # pH is not known
        if ABT is not None:
            H = calculate_H(Ks=Ks, alphaB=alphaB, ABT=ABT, ABO3=ABO3, ABO4=ABO4)
            pHtot = -np.log10(H)
        else:
            raise ValueError('ABT and one of ABO3 or ABO4 must be specified if pH is missing.')
    
    if ABO3 is None:
        ABO3 = calculate_ABO3(H=H, Ks=Ks, ABT=ABT, alphaB=alphaB)
    if ABO4 is None:
        ABO4 = calculate_ABO4(H=H, Ks=Ks, ABT=ABT, alphaB=alphaB)
    
    return Bunch({
        'pHtot': pHtot,
        'ABT': ABT,
        'ABO4': ABO4,
        'ABO3': ABO3,
        'H': H
    })

# Wrapper functions using delta values
def calculate_pH(Ks, d11BT, d11B4, epsilon=get_epsilonB()):
    """
    Calculates pH on the total scale

    Parameters
    ----------
    Ks : Bunch (dictionary with . access)
        bunch containing the boron speciation constant KB
    d11BT : float or array-like
        isotope ratio 11B/10B in total boron - delta units
    d11B4 : float or array-like
        isotope ratio 11B/10B in BO4 - delta units, in ‰
    epsilon : float or array-like
        fractionation factor between BO3 and BO4, in ‰

    Returns
    ----------
    array-like
        pH on the total scale
    """
    ABO4 = d11_to_A11(d11B4)
    ABT = d11_to_A11(d11BT)
    alphaB = epsilon_to_alpha(epsilon)

    return -np.log10(calculate_H(Ks,alphaB,ABT,ABO4))

def calculate_pKB(pH, d11BT, d11B4, epsilonB=get_epsilonB()):
    """
    Calculate stoichiometric equilibrium constant for boron with delta inputs

    Parameters
    ----------
    pH : array-like
        pH on the total scale
    d11BT : array-like
        The isotope ratio of 11B in total B in delta units, in ‰
    d11B4 : array-like
        The isotope ratio of 11B in borate ion (B(OH)4) in delta units, in ‰
    epsilonB : array-like
        The fractionation factor between B(OH)3 and B(OH)4- as delta units, in ‰

    Returns
    -------
    array-like
        The stoichiometric equilibrium constant for boron (KB)
    """
    ABO4 = d11_to_A11(d11B4)
    ABT = d11_to_A11(d11BT)
    H = 10.0**-pH

    alphaB = epsilon_to_alpha(epsilonB)

    return -np.log10(calculate_KB(H,alphaB,ABT,ABO4))

def calculate_d11BT(pH, KB, d11B4, epsilonB=get_epsilonB()):
    """
    Calcluates the isotope ratio of total boron in delta units

    Parameters
    ----------
    pH : float or array-like
        pH on the total scale
    KB : Bunch (dictionary with . access)
        bunch containing the boron speciation constant KB
    d11B4 : float or array-like
        isotope ratio 11B/10B in BO4 - delta units, in ‰
    epsilonB : float or array-like
        fractionation factor between BO3 and BO4, units of ‰

    Returns
    -------
    array-like
        The isotope ratio 11B/10B in BT - delta units (d11BT), in ‰
    """
    ABO4 = d11_to_A11(d11B4)
    alphaB = epsilon_to_alpha(epsilonB)
    H = 10.0**-pH
    return A11_to_d11(calculate_ABT(H,KB,alphaB,ABO4))

def calculate_d11B4(pH, KB, d11BT, epsilonB=get_epsilonB()):
    """
    Calculates the isotope ratio of borate ion in delta units

    Parameters
    ----------
    pH : float or array-like
        pH on the total scale
    KB : Bunch (dictionary with . access)
        bunch containing the boron speciation constant KB
    d11BT : float or array-like
        isotope ratio 11B/10B in total boron - delta units, in ‰
    epsilonB : float or array-like
        fractionation factor between BO3 and BO4, units of ‰
    
    Returns
    -------
    array-like
        The isotope ratio 11B/10B in BO4 - delta units, in ‰
    """
    ABOT = d11_to_A11(d11BT)
    alphaB = epsilon_to_alpha(epsilonB)

    return A11_to_d11(calculate_ABO4(10.0**-pH,KB,ABOT,alphaB))

def calculate_epsilon(pH, KB, d11BT, d11B4):
    """
    Returns isotope ratio of borate ion in delta units

    Parameters
    ----------
    pH : float or array-like
        pH on the total scale
    KB : Bunch (dictionary with . access)
        bunch containing the boron speciation constant KB
    d11BT : float or array-like
        isotope ratio 11B/10B in total boron - delta units, in ‰
    d11B4 : float or array-like
        isotope ratio 11B/10B in borate ion (B(OH)4) - delta units, in ‰
        
    Returns
    -------
    array-like
        fractionation factor between BO3 and BO4 in delta units (epsilon, in ‰)
    """
    ABO4 = d11_to_A11(d11B4)
    ABT = d11_to_A11(d11BT)
    H = 10.0**-pH

    alphaB = calculate_alpha_ABO4(H,KB,ABT,ABO4)

    return alpha_to_epsilon(alphaB)



import unittest
import cbsyst as cb

from tests import check_vals
ref = check_vals.carbon_ref

class TestInputOutput(unittest.TestCase):
    """Test internal consistency of input/output condition calculations"""

    def test_Csys(self):
        
        with self.subTest(msg='Temperature Effect'):      
            c1 = cb.Csys(pHtot=8.1, TA=2300, T_in=20, T_out=30)
            c2 = cb.Csys(pHtot=c1.pHtot, TA=2300, T_in=30, T_out=20)
            self.assertAlmostEqual(c1.pHtot_in, c2.pHtot.item(), places=6)
            
        with self.subTest(msg='Salinity Effect'):      
            c1 = cb.Csys(pHtot=8.1, TA=2300, S_in=28.2, S_out=38.1)
            c2 = cb.Csys(pHtot=c1.pHtot, TA=2300, S_in=38.1, S_out=28.2)
            self.assertAlmostEqual(c1.pHtot_in, c2.pHtot.item(), places=6)
            
        with self.subTest(msg='Pressure Effect'):      
            c1 = cb.Csys(pHtot=8.1, TA=2300, P_in=0, P_out=400)
            c2 = cb.Csys(pHtot=c1.pHtot, TA=2300, P_in=400, P_out=0)
            self.assertAlmostEqual(c1.pHtot_in, c2.pHtot.item(), places=6)
            

    def test_CBsys(self):
        
        with self.subTest(msg='Temperature Effect'):      
            c1 = cb.CBsys(pHtot=8.1, TA=2300, T_in=20, T_out=30, dBT=39.4)
            c2 = cb.CBsys(pHtot=c1.pHtot, TA=2300, T_in=30, T_out=20, dBT=39.4)
            self.assertAlmostEqual(c1.pHtot_in.item(), c2.pHtot.item(), places=6)
            
        with self.subTest(msg='Salinity Effect'):      
            c1 = cb.CBsys(pHtot=8.1, TA=2300, S_in=28.2, S_out=38.1, dBT=39.4)
            c2 = cb.CBsys(pHtot=c1.pHtot, TA=2300, S_in=38.1, S_out=28.2, dBT=39.4)
            self.assertAlmostEqual(c1.pHtot_in.item(), c2.pHtot.item(), places=6)
            
        with self.subTest(msg='Pressure Effect'):      
            c1 = cb.CBsys(pHtot=8.1, TA=2300, P_in=0, P_out=400, dBT=39.4)
            c2 = cb.CBsys(pHtot=c1.pHtot, TA=2300, P_in=400, P_out=0, dBT=39.4)
            self.assertAlmostEqual(c1.pHtot_in.item(), c2.pHtot.item(), places=6)
            


import numpy as np
from cbsyst.helpers import Bunch

def chiB_calc(H, Ks):
    return 1 / (1 + Ks.KB / H)

# B conc fns
def BT_BO3(BT, BO3, Ks):
    """
    Returns H
    """
    return Ks.KB / (BT / BO3 - 1)


def BT_BO4(BT, BO4, Ks):
    """
    Returns H
    """
    return Ks.KB * (BT / BO4 - 1)


def pH_BO3(pH, BO3, Ks):
    """
    Returns BT
    """
    H = 10.0**-pH
    return BO3 * (1 + Ks.KB / H)


def pH_BO4(pH, BO4, Ks):
    """
    Returns BT
    """
    H = 10.0**-pH
    return BO4 * (1 + H / Ks.KB)


def cBO4(BT, H, Ks):
    return BT / (1 + H / Ks.KB)


def cBO3(BT, H, Ks):
    return BT / (1 + Ks.KB / H)

def calc_B_species(pHtot=None, BT=None, BO3=None, BO4=None, Ks=None, **kwargs):
    # B system calculations
    if pHtot is not None and BT is not None:
        H = 10.0**-pHtot
    elif BT is not None and BO3 is not None:
        H = BT_BO3(BT, BO3, Ks)
    elif BT is not None and BO4 is not None:
        H = BT_BO4(BT, BO4, Ks)
    elif BO3 is not None and BO4 is not None:
        BT = BO3 + BO3
        H = BT_BO3(BT, BO3, Ks)
    elif pHtot is not None and BO3 is not None:
        H = 10.0**-pHtot
        BT = pH_BO3(pHtot, BO3, Ks)
    elif pHtot is not None and BO4 is not None:
        H = 10.0**-pHtot
        BT = pH_BO4(pHtot, BO4, Ks)

    # The above makes sure that BT and H are known,
    # this next bit calculates all the missing species
    # from BT and H.

    if BO3 is None:
        BO3 = cBO3(BT, H, Ks)
    if BO4 is None:
        BO4 = cBO4(BT, H, Ks)
    if pHtot is None:
        pHtot = np.array(-np.log10(H), ndmin=1)

    return Bunch({"Htot": pHtot, "H": H, "BT": BT, "BO3": BO3, "BO4": BO4})



import unittest
import os
import pandas as pd
import numpy as np
import cbsyst.carbon as cf
import cbsyst.boron as bf
import cbsyst.boron_isotopes as bif
from cbsyst.cbsyst import Bsys, CBsys, ABsys

np.random.seed(42)

n = 101
pHtot = np.random.uniform(7.6, 8.8, n)
DIC = np.random.uniform(1900, 2200, n)
BT = np.random.uniform(350, 450, n)
dBT = np.random.uniform(35, 45, n)
T = np.random.uniform(15, 35, n)
S = np.random.uniform(30, 40, n)

test = CBsys(pHtot=pHtot, dBT=dBT, BT=BT, DIC=DIC, T_in=T, S_in=S)


class ReferenceDataTestCase(unittest.TestCase):
    """Test boron idotopes"""

    def test_Bisotopes(self):
        check_A = ABsys(dBO4=test.dBO4, dBT=test.dBT, T_in=T, S_in=S)
        
        self.assertIsNone(np.testing.assert_allclose(test.pHtot, check_A.pHtot, rtol=1e-10))
        
        check_B = Bsys(dBO4=test.dBO4, dBT=test.dBT, BT=test.BT, T_in=T, S_in=S)
    
        self.assertIsNone(np.testing.assert_allclose(test.pHtot, check_B.pHtot, rtol=1e-10))

        check_CB = CBsys(dBO4=test.dBO4, dBT=test.dBT, DIC=test.DIC, BT=test.BT, T_in=T, S_in=S)

        self.assertIsNone(np.testing.assert_allclose(test.pHtot, check_CB.pHtot, rtol=1e-10))



import unittest
import os
import pandas as pd
import numpy as np
import cbsyst.carbon as cf
import cbsyst.boron as bf
import cbsyst.boron_isotopes as bif
from cbsyst.cbsyst import Csys, CBsys
from cbsyst.helpers import Bunch
from tests.test_data.GLODAP_data.get_GLODAP_data import get_GLODAP

class ReferenceDataTestCase(unittest.TestCase):
    """Test `yt` against reference data."""

    def test_Bockmon_Data_Csys(self):
        # Measured data from paper
        batch_A = {"S": 33.190, "TA": 2215.08, "DIC": 2015.72, "pH": 7.8796}

        batch_B = {"S": 33.186, "TA": 2216.26, "DIC": 2141.94, "pH": 7.5541}

        pH = np.array([batch_A["pH"], batch_B["pH"]])
        TA = np.array([batch_A["TA"], batch_B["TA"]])
        DIC = np.array([batch_A["DIC"], batch_B["DIC"]])
        S = np.array([batch_A["S"], batch_B["S"]])
        BT = 433.0

        # Csys calculations
        # TA from pH and DIC
        cTA = Csys(pHtot=pH, DIC=DIC, BT=BT, S_in=S)
        # Calculate % differences from measured
        dTA = 100 * (TA - cTA.TA) / TA

        self.assertLess(max(abs(dTA)), 0.2, msg="TA from DIC and pH")

        # pH from TA and DIC
        cpH = Csys(DIC=DIC, TA=TA, BT=BT, S_in=S)
        # Calculate % differences from measured
        dpH = 100 * (pH - cpH.pHtot) / pH

        self.assertLess(max(abs(dpH)), 0.2, msg="pH from TA and DIC")

        # DIC from pH and TA
        cDIC = Csys(pHtot=pH, TA=TA, BT=BT, S_in=S)
        # Calculate % differences from measured
        dDIC = 100 * (DIC - cDIC.DIC) / DIC

        self.assertLess(max(abs(dDIC)), 0.2, msg="DIC from TA and pH")

        return

    def test_Bockmon_Data_CBsys(self):
        # Measured data from paper
        batch_A = {"S": 33.190, "TA": 2215.08, "DIC": 2015.72, "pH": 7.8796}

        batch_B = {"S": 33.186, "TA": 2216.26, "DIC": 2141.94, "pH": 7.5541}

        pH = np.array([batch_A["pH"], batch_B["pH"]])
        TA = np.array([batch_A["TA"], batch_B["TA"]])
        DIC = np.array([batch_A["DIC"], batch_B["DIC"]])
        S = np.array([batch_A["S"], batch_B["S"]])
        BT = 433.0

        # Csys calculations
        # TA from pH and DIC
        cTA = CBsys(pHtot=pH, DIC=DIC, BT=BT, S_in=S)
        # Calculate % differences from measured
        dTA = 100 * (TA - cTA.TA) / TA

        self.assertLess(max(abs(dTA)), 0.2, msg="TA from DIC and pH")

        # pH from TA and DIC
        cpH = CBsys(DIC=DIC, TA=TA, BT=BT, S_in=S)
        # Calculate % differences from measured
        dpH = 100 * (pH - cpH.pHtot) / pH

        self.assertLess(max(abs(dpH)), 0.2, msg="pH from TA and DIC")

        # DIC from pH and TA
        cDIC = CBsys(pHtot=pH, TA=TA, BT=BT, S_in=S)
        # Calculate % differences from measured
        dDIC = 100 * (DIC - cDIC.DIC) / DIC

        self.assertLess(max(abs(dDIC)), 0.2, msg="DIC from TA and pH")

        return

    def test_Lueker_Data_Csys(self):
        """
        Need to incorporate nutrients!
        """
        ld = pd.read_csv(
            "tests/test_data/Lueker2000/Lueker2000_Table3.csv", comment="#"
        )

        # Calculate using cbsys
        # TA from DIC and fCO2
        cTA = Csys(
            DIC=ld.DIC.values,
            fCO2=ld.fCO2.values,
            T_in=ld.Temp.values,
            S_in=ld.Sal.values,
        )
        dTA = ld.TA - cTA.TA
        dTA_median = np.median(dTA)
        dTA_pc95 = np.percentile(dTA, [2.5, 97.5])
        self.assertLessEqual(abs(dTA_median), 2.5, msg="TA Offset <= 2.5")
        self.assertTrue(all(abs(dTA_pc95 - dTA_median) <= 16), msg="TA 95% Conf <= 16")

        # fCO2 from TA and DIC
        cfCO2 = Csys(
            TA=ld.TA.values, DIC=ld.DIC.values, T_in=ld.Temp.values, S_in=ld.Sal.values
        )
        dfCO2 = ld.fCO2 - cfCO2.fCO2
        dfCO2_median = np.median(dfCO2)
        # dfCO2_pc95 = np.percentile(dfCO2, [2.5, 97.5])
        dfCO2_percent_offset = 100 * dfCO2 / ld.fCO2
        self.assertLessEqual(dfCO2_median, 2.5, msg="fCO2 Offset <= 2.5")
        self.assertLessEqual(np.std(dfCO2_percent_offset), 3, msg="fCO2 STD within 3%")
        # print(dfCO2_pc95)
        # self.assertTrue(all(abs(dfCO2_pc95) <= 70), msg='fCO2 95% Conc <= 70')

        # DIC from TA and fCO2
        cDIC = Csys(
            TA=ld.TA.values,
            fCO2=ld.fCO2.values,
            T_in=ld.Temp.values,
            S_in=ld.Sal.values,
        )
        dDIC = ld.DIC - cDIC.DIC
        dDIC_median = np.median(dDIC)
        dDIC_pc95 = np.percentile(dDIC, [2.5, 97.5])
        self.assertLessEqual(abs(dDIC_median), 2, msg="DIC Offset <= 2")
        self.assertTrue(all(abs(dDIC_pc95) <= 15), msg="DIC 95% Conc <= 15")

        return

    def test_Lueker_Data_CBsys(self):
        """
        Need to incorporate nutrients!
        """
        ld = pd.read_csv(
            "tests/test_data/Lueker2000/Lueker2000_Table3.csv", comment="#"
        )

        # Calculate using cbsys
        # TA from DIC and fCO2
        cTA = CBsys(
            DIC=ld.DIC.values,
            fCO2=ld.fCO2.values,
            T_in=ld.Temp.values,
            S_in=ld.Sal.values,
        )
        dTA = ld.TA - cTA.TA
        dTA_median = np.median(dTA)
        dTA_pc95 = np.percentile(dTA, [2.5, 97.5])
        self.assertLessEqual(abs(dTA_median), 2.5, msg="TA Offset <= 2.5")
        self.assertTrue(all(abs(dTA_pc95 - dTA_median) <= 16), msg="TA 95% Conf <= 16")

        # fCO2 from TA and DIC
        cfCO2 = CBsys(
            TA=ld.TA.values, DIC=ld.DIC.values, T_in=ld.Temp.values, S_in=ld.Sal.values
        )
        dfCO2 = ld.fCO2 - cfCO2.fCO2
        dfCO2_median = np.median(dfCO2)
        # dfCO2_pc95 = np.percentile(dfCO2, [2.5, 97.5])
        dfCO2_percent_offset = 100 * dfCO2 / ld.fCO2
        self.assertLessEqual(dfCO2_median, 2.5, msg="fCO2 Offset <= 2.5")
        self.assertLessEqual(np.std(dfCO2_percent_offset), 3, msg="fCO2 STD within 3%")
        # print(dfCO2_pc95)
        # self.assertTrue(all(abs(dfCO2_pc95) <= 70), msg='fCO2 95% Conc <= 70')

        # DIC from TA and fCO2
        cDIC = CBsys(
            TA=ld.TA.values,
            fCO2=ld.fCO2.values,
            T_in=ld.Temp.values,
            S_in=ld.Sal.values,
        )
        dDIC = ld.DIC - cDIC.DIC
        dDIC_median = np.median(dDIC)
        dDIC_pc95 = np.percentile(dDIC, [2.5, 97.5])
        self.assertLessEqual(abs(dDIC_median), 2, msg="DIC Offset <= 2")
        self.assertTrue(all(abs(dDIC_pc95) <= 15), msg="DIC 95% Conc <= 15")

        return

    def test_GLODAPv2_Csys(self):
        """
        Test Csys against GLODAP data (n = 83,030).

        Check median offsets are within acceptable limits.
        Check 95% confidence of residuals are within acceptable limits.
        """
        # load GLODAP data
        gd = pd.read_csv("tests/test_data/GLODAP_data/GLODAPv2_pH_DIC_ALK_subset.csv")
        gd.dropna(
            subset=[
                "phtsinsitutp",
                "temperature",
                "salinity",
                "tco2",
                "talk",
                "pressure",
                "phosphate",
                "silicate",
            ],
            inplace=True,
        )
        gd.pressure /= 10  # convert pressure to bar

        # set negative nutrient values to zero
        gd.phosphate[gd.phosphate < 0] = 0
        gd.silicate[gd.silicate < 0] = 0
        
        # exclude weird cruise 270 data
        gd = gd.loc[gd.cruise != 270]

        # calculate pH from TA and DIC
        cpH = Csys(
            TA=gd.talk,
            DIC=gd.tco2,
            T_in=gd.temperature,
            S_in=gd.salinity,
            P_in=gd.pressure,
            PT=gd.phosphate,
            SiT=gd.silicate,
        )
        pH_resid = gd.phtsinsitutp - cpH.pHtot
        pH_median = np.median(pH_resid)
        pH_pc95 = np.percentile(pH_resid, [2.5, 97.5])

        self.assertLessEqual(abs(pH_median), 0.005, msg="pH Offset <= 0.005")
        self.assertTrue(all(abs(pH_pc95) <= 0.05), msg="pH 95% Conf <= 0.05")

        # calculate TA from pH and DIC
        cTA = Csys(
            pHtot=gd.phtsinsitutp,
            DIC=gd.tco2,
            T_in=gd.temperature,
            S_in=gd.salinity,
            P_in=gd.pressure,
            PT=gd.phosphate,
            SiT=gd.silicate,
            BT=415.7,
        )
        TA_resid = gd.talk - cTA.TA
        TA_median = np.median(TA_resid)
        TA_pc95 = np.percentile(TA_resid, [2.5, 97.5])

        print(abs(TA_median))
        self.assertLessEqual(abs(TA_median), 0.6, msg="TA Offset <= 0.6")
        self.assertTrue(all(abs(TA_pc95) < 13), msg="TA 95% Conf <= 13")

        # calculate DIC from TA and pH
        cDIC = Csys(
            pHtot=gd.phtsinsitutp,
            TA=gd.talk,
            T_in=gd.temperature,
            S_in=gd.salinity,
            P_in=gd.pressure,
            PT=gd.phosphate,
            SiT=gd.silicate,
            BT=415.7,
        )
        DIC_resid = gd.tco2 - cDIC.DIC
        DIC_median = np.median(DIC_resid)
        DIC_pc95 = np.percentile(DIC_resid, [2.5, 97.5])

        self.assertLessEqual(abs(DIC_median), 0.6, msg="DIC Offset <= 0.6")
        self.assertTrue(all(abs(DIC_pc95) < 13), msg="DIC 95% Conf <= 13")

        return

    def test_GLODAPv2_CBsys(self):
        """
        Test Csys against GLODAP data (n = 83,030).

        Check median offsets are within acceptable limits.
        Check 95% confidence of residuals are within acceptable limits.
        """
        if not os.path.exists(
            "tests/test_data/GLODAP_data/GLODAPv2_pH_DIC_ALK_subset.csv"
        ):
            get_GLODAP(path="tests/test_data/GLODAP_data/", leave_zip=True)

        # load GLODAP data
        gd = pd.read_csv("tests/test_data/GLODAP_data/GLODAPv2_pH_DIC_ALK_subset.csv")
        gd.dropna(
            subset=[
                "phtsinsitutp",
                "temperature",
                "salinity",
                "tco2",
                "talk",
                "pressure",
                "phosphate",
                "silicate",
            ],
            inplace=True,
        )
        gd.pressure /= 10  # convert pressure to bar
        
        # set negative nutrient values to zero
        gd.phosphate[gd.phosphate < 0] = 0
        gd.silicate[gd.silicate < 0] = 0
        
        # exclude weird cruise 270 data
        gd = gd.loc[gd.cruise != 270]

        # calculate pH from TA and DIC
        cpH = CBsys(
            TA=gd.talk,
            DIC=gd.tco2,
            T_in=gd.temperature,
            S_in=gd.salinity,
            P_in=gd.pressure,
            PT=gd.phosphate,
            SiT=gd.silicate,
            BT=415.7,
        )
        pH_resid = gd.phtsinsitutp - cpH.pHtot
        pH_median = np.median(pH_resid)
        pH_pc95 = np.percentile(pH_resid, [2.5, 97.5])

        self.assertLessEqual(abs(pH_median), 0.005, msg="pH Offset <= 0.005")
        self.assertTrue(all(abs(pH_pc95) <= 0.05), msg="pH 95% Conf <= 0.05")

        # calculate TA from pH and DIC
        cTA = CBsys(
            pHtot=gd.phtsinsitutp,
            DIC=gd.tco2,
            T_in=gd.temperature,
            S_in=gd.salinity,
            P_in=gd.pressure,
            PT=gd.phosphate,
            SiT=gd.silicate,
            BT=415.7,
        )
        TA_resid = gd.talk - cTA.TA
        TA_median = np.median(TA_resid)
        TA_pc95 = np.percentile(TA_resid, [2.5, 97.5])

        self.assertLessEqual(abs(TA_median), 1, msg="TA Offset <= 1")
        self.assertTrue(all(abs(TA_pc95) < 15), msg="TA 95% Conf <= 15")

        # calculate DIC from TA and pH
        cDIC = CBsys(
            pHtot=gd.phtsinsitutp,
            TA=gd.talk,
            T_in=gd.temperature,
            S_in=gd.salinity,
            P_in=gd.pressure,
            PT=gd.phosphate,
            SiT=gd.silicate,
            BT=415.7,
        )
        DIC_resid = gd.tco2 - cDIC.DIC
        DIC_median = np.median(DIC_resid)
        DIC_pc95 = np.percentile(DIC_resid, [2.5, 97.5])

        self.assertLessEqual(abs(DIC_median), 1, msg="DIC Offset <= 1")
        self.assertTrue(all(abs(DIC_pc95) < 15), msg="DIC 95% Conf <= 15")



import os
import tarfile
import urllib.request as ureq

import pandas as pd
import numpy as np
from scipy.interpolate import RegularGridInterpolator

from tqdm import tqdm


# Status bar
class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


def get_GLODAP_WOA(save=True):
    # download GLODAP data if not present
    if not os.path.exists("./GLODAP_data/"):
        print("Downloading GLODAP data...")
        os.mkdir("./GLODAP_data")
        # download data
        GLODAP_url = "http://cdiac.ornl.gov/ftp/oceans/GLODAP_Gridded_Data/{:}"
        GLODAP_files = ["Alk.tar.Z", "TCO2.tar.Z"]
        for f in GLODAP_files:
            with TqdmUpTo(
                unit="B", unit_scale=True, miniters=1, desc="GLODAP {:}".format(f)
            ) as t:
                ureq.urlretrieve(
                    GLODAP_url.format(f),
                    "./GLODAP_data/{:}".format(f),
                    reporthook=t.update_to,
                )

    # download WOA data if not present
    if not os.path.exists("./WOA2009_data/"):
        print("Downloading WOA2009 data...")
        os.mkdir("./WOA2009_data")
        # directory: parameter_key
        WOA_url = "ftp://ftp.nodc.noaa.gov/pub/woa/WOA09/DATA/{:}/csv/{:}_00_1d.tar.gz"
        WOA_params = {
            "salinity": "s",
            "temperature": "t",
            "nitrate": "n",
            "phosphate": "p",
            "silicate": "i",
        }
        # download data
        for d, p in WOA_params.items():
            url = WOA_url.format(d, p)
            fname = "./WOA2009_data/" + url.split("/")[-1]

            with TqdmUpTo(
                unit="B", unit_scale=True, miniters=1, desc="WOA2009 {:}".format(d)
            ) as t:
                ureq.urlretrieve(url, fname, reporthook=t.update_to)

    # Load Data
    # ---------

    # GLODAP
    # ++++++
    print("Reading GLODAP data...")

    # Uncompress files (Python can't read them...)
    try:
        for f in [z for z in os.listdir("GLODAP_data/") if "tar.Z" in z]:
            os.system("uncompress ./GLODAP_data/" + f)
    except:
        raise RuntimeError(
            "Cannot unzip GLODAP data files. Perhaps you are\non Windows? If so, please unzip them manually (extension shoudl be .tar)\nthen try again."
        )

    # load GLODAP data
    glodap = {}
    for f in os.listdir("GLODAP_data/"):
        var = f.split(".")[0]
        tar = tarfile.open("GLODAP_data/{:}.tar".format(var))
        tmp = np.genfromtxt(tar.extractfile(var + "/" + var + ".data"), delimiter=",")
        # reshape to 3D array
        glodap[var.lower()] = np.reshape(tmp, (33, 360, 180))

    # axes
    lat = np.genfromtxt(tar.extractfile(var + "/" + "Lat.centers"), delimiter=",")
    lon = np.genfromtxt(tar.extractfile(var + "/" + "Long.centers"), delimiter=",")
    dep = np.genfromtxt(tar.extractfile(var + "/" + "Depth.centers"), delimiter=",")
    glon, gdep, glat = np.meshgrid(lon, dep, lat)

    # WOA-2009
    # ++++++++
    # set up column headins and params
    print("Reading WOA2009 data...")

    columns = [
        "lat",
        "lon",
        "depth",
        "obj_mean",
        "mean",
        "std",
        "se",
        "obj_mean_mean",
        "obj_mean_ann_mean",
        "ngrids",
        "ndata",
    ]
    params = ["s", "t", "n", "p", "i"]
    pdict = {
        k: v
        for k, v in zip(params, ["sal", "temp", "nitrate", "phosphate", "silicate"])
    }

    # read in all data
    cdat = []
    for p in params:
        tar = tarfile.open("WOA2009_data/" + p + "_00_1d.tar.gz")

        dat = []
        for f in tar.getnames():
            dat.append(pd.read_csv(tar.extractfile(f), names=columns, comment="#"))
        dat = pd.concat(dat)
        dat.loc[:, "var"] = p

        cdat.append(dat)
    cdat = pd.concat(cdat).loc[:, ["lat", "lon", "depth", "obj_mean", "var"]]

    # recaset into 3D arrays
    woa = pd.pivot_table(cdat, index=["lat", "lon", "depth"], columns="var")
    woa.columns = woa.columns.droplevel()
    woa.columns = [pdict[c] for c in woa.columns]
    params = [pdict[p] for p in params]

    shape = list(map(len, woa.index.levels))

    gwoa = {}
    for p in params:
        # create empy
        gwoa[p] = np.full(shape, np.nan)
        # fill with variables based on MultiIndex levels
        gwoa[p][woa.index.labels] = woa.loc[:, p].values.flat

    # axes
    woa_lat = woa.index.levels[0].values
    woa_lon = woa.index.levels[1].values
    woa_dep = woa.index.levels[2].values

    # create interpolators for use later.
    interp = {}
    for p in params:
        interp[p] = RegularGridInterpolator((woa_lat, woa_lon, woa_dep), gwoa[p])

    # Combine Datasets
    # ++++++++++++++++
    # dataframe of glodap data
    print("Generating GLODAP-WOA dataset...")
    glodap_woa = pd.DataFrame(
        {"lat": glat.ravel(), "lon": glon.ravel(), "dep": gdep.ravel()}
    )
    for k, v in glodap.items():
        glodap_woa.loc[:, k] = v.ravel()

    # replace -999 with nan
    glodap_woa.replace(-999, np.nan, inplace=True)
    # drop nan
    glodap_woa.dropna(inplace=True, subset=glodap.keys())

    # interpolate WOA parameters onto GLODAP grid
    print("Interpolating WOA2009 onto GLODAP grid...")
    for p in params:
        glodap_woa.loc[:, p] = interp[p](
            glodap_woa.loc[:, ["lat", "lon", "dep"]].values
        )

    # drop failed interpolation values
    glodap_woa.dropna(inplace=True, subset=params)

    # very approx pressure estimate...
    glodap_woa.loc[:, "pres"] = glodap_woa.loc[:, "dep"] / 10

    # reset index
    glodap_woa.reset_index(inplace=True, drop=True)

    # save data
    if save:
        print("Saving...")
        glodap_woa.to_pickle("glodap_woa.pkl")
        return
    else:
        return glodap_woa


if __name__ == "__main__":
    print("\n*************************************************")
    print("Making GLODAP-WOA dataset (after Orr et al, 2015)")
    print("*************************************************")
    get_GLODAP_WOA()
    print("Done.")
    print()
    print("   Saved as: ./glodap_woa.pkl")
    print("             (read with pandas.read_pickle())")
    print("*************************************************\n")



import unittest
from cbsyst import boron as bf

from tests import check_vals
ref = check_vals.boron_ref
class BoronFunctions(unittest.TestCase):
    """Test B concentration and speciation functions"""

    def test_boron(self):

        # Speciation
        with self.subTest(msg='BT_BO3'):
            self.assertAlmostEqual(
                bf.BT_BO3(BT=ref.BT, BO3=ref.BO3, Ks=ref.Ks), ref.H, places=6
            )

        with self.subTest(msg='BT_BO4'):
            self.assertAlmostEqual(
                bf.BT_BO4(BT=ref.BT, BO4=ref.BO4, Ks=ref.Ks), ref.H, places=6
            )

        with self.subTest(msg='pH_BO3'):
            self.assertAlmostEqual(
                bf.pH_BO3(ref.pHtot, BO3=ref.BO3, Ks=ref.Ks), ref.BT, places=6
            )

        with self.subTest(msg='pH_BO3'):
            self.assertAlmostEqual(
                bf.pH_BO4(ref.pHtot, BO4=ref.BO4, Ks=ref.Ks), ref.BT, places=6
            )

        with self.subTest(msg='cBO3'):
            self.assertAlmostEqual(
                bf.cBO3(BT=ref.BT, H=ref.H, Ks=ref.Ks), ref.BO3, places=6
            )

        with self.subTest(msg='cBO4'):
            self.assertAlmostEqual(
                bf.cBO4(BT=ref.BT, H=ref.H, Ks=ref.Ks), ref.BO4, places=6
            )

        with self.subTest(msg='chiB_calc'):
            self.assertEqual(
                bf.chiB_calc(ref.H, Ks=ref.Ks), 1 / (1 + ref.Ks.KB / ref.H)
                )



import scipy.optimize as opt
import numpy as np
from cbsyst.helpers import noms, cast_array, Bunch, maxShape, calc_fH

def _zero_wrapper(ps, fn, bounds=(10 ** -14, 10 ** -1)):
    """
    Wrapper to handle zero finders.
    """
    try:
        return opt.brentq(fn, *bounds, args=tuple(ps), xtol=1e-16)
        # brentq is ~100 times faster.
    except ValueError:
        return opt.fsolve(fn, 1, args=tuple(ps))[0]
        # but can be fragile if limits aren't right.


# Function types
# Zero-finders: 2-5, 10-15
# Algebraic: 1, 6-9


# Zeebe & Wolf-Gladrow, Appendix B
# 1. CO2 and pH given
def CO2_pH(CO2, pH, Ks):
    """
    Returns DIC
    """
    h = 10.0**-pH
    return CO2 * (1 + Ks.K1 / h + Ks.K1 * Ks.K2 / h ** 2)


# 2. CO2 and HCO3 given
def CO2_HCO3(CO2, HCO3, Ks):
    """
    Returns H
    """
    CO2, HCO3 = noms(CO2, HCO3)  # get nominal values of inputs
    par = cast_array(CO2, HCO3, Ks.K1, Ks.K2)  # cast parameters into array
    shape = maxShape(CO2, HCO3, Ks.K1, Ks.K2)  # get shape of output


    return np.apply_along_axis(_zero_wrapper, 0, par, fn=zero_CO2_HCO3).reshape(shape)


def zero_CO2_HCO3(h, CO2, HCO3, K1, K2):
    # Roots: two negative, one positive - use positive.
    LH = CO2 * (h ** 2 + K1 * h + K1 * K2)
    RH = HCO3 * (h ** 2 + h ** 3 / K1 + K2 * h)
    return LH - RH


# 3. CO2 and CO3
def CO2_CO3(CO2, CO3, Ks):
    """
    Returns H
    """
    CO2, CO3 = noms(CO2, CO3)
    par = cast_array(CO2, CO3, Ks.K1, Ks.K2)  # cast parameters into array
    shape = maxShape(CO2, CO3, Ks.K1, Ks.K2)  # get shape of output


    return np.apply_along_axis(_zero_wrapper, 0, par, fn=zero_CO2_CO3).reshape(shape)


def zero_CO2_CO3(h, CO2, CO3, K1, K2):
    # Roots: one positive, three negative. Use positive.
    LH = CO2 * (h ** 2 + K1 * h + K1 * K2)
    RH = CO3 * (h ** 2 + h ** 3 / K2 + h ** 4 / (K1 * K2))
    return LH - RH


# 4. CO2 and TA
def CO2_TA(CO2, TA, BT, PT, SiT, ST, FT, Ks):
    """
    Returns pH

    Taken from matlab CO2SYS
    """
    
    fCO2 = CO2 / Ks.K0
    L = maxShape(TA, CO2, BT, PT, SiT, ST, FT, Ks.K1)
    if len(L) == 0:
        L = (1,)
        
    pHguess = 8.0
    pHtol = 0.0000001
    pHx = np.full(L, pHguess)
    deltapH = np.array(pHtol + 1, ndmin=1)
    ln10 = np.log(10)

    while np.any(abs(deltapH) > pHtol):
        H = 10 ** -pHx
        HCO3 = Ks.K0 * Ks.K1 * fCO2 / H
        CO3 = Ks.K0 * Ks.K1 * Ks.K2 * fCO2 / H ** 2
        CAlk = HCO3 + 2 * CO3
        BAlk = BT * Ks.KB / (Ks.KB + H)
        OH = Ks.KW / H
        PhosTop = Ks.KP1 * Ks.KP2 * H + 2 * Ks.KP1 * Ks.KP2 * Ks.KP3 - H ** 3
        PhosBot = (
            H ** 3 + Ks.KP1 * H ** 2 + Ks.KP1 * Ks.KP2 * H + Ks.KP1 * Ks.KP2 * Ks.KP3
        )
        PAlk = PT * PhosTop / PhosBot
        SiAlk = SiT * Ks.KSi / (Ks.KSi + H)
        # positive
        Hfree = H / (1 + ST / Ks.KS)
        HSO4 = ST / (1 + Ks.KS / Hfree)
        HF = FT / (1 + Ks.KF / Hfree)

        Residual = TA - CAlk - BAlk - OH - PAlk - SiAlk + Hfree + HSO4 + HF
        Slope = ln10 * (HCO3 + 4.0 * CO3 + BAlk * H / (Ks.KB + H) + OH + H)
        deltapH = Residual / Slope

        while np.any(abs(deltapH) > 1):
            FF = abs(deltapH) > 1
            deltapH[FF] = deltapH[FF] / 2

        pHx += deltapH

    return pHx


# 5. CO2 and DIC
def CO2_DIC(CO2, DIC, Ks):
    """
    Returns H
    """
    CO2, DIC = noms(CO2, DIC)  # get nominal values of inputs
    par = cast_array(CO2, DIC, Ks.K1, Ks.K2)  # cast parameters into array
    shape = maxShape(CO2, DIC, Ks.K1, Ks.K2)  # get shape of output

    return np.apply_along_axis(_zero_wrapper, 0, par, fn=zero_CO2_DIC).reshape(shape)


def zero_CO2_DIC(h, CO2, DIC, K1, K2):
    # Roots: one positive, one negative. Use positive.
    LH = DIC * h ** 2
    RH = CO2 * (h ** 2 + K1 * h + K1 * K2)
    return LH - RH


# 6. pH and HCO3
def pH_HCO3(pH, HCO3, Ks):
    """
    Returns DIC
    """
    h = 10.0**-pH
    return HCO3 * (1 + h / Ks.K1 + Ks.K2 / h)


# 7. pH and CO3
def pH_CO3(pH, CO3, Ks):
    """
    Returns DIC
    """
    h = 10.0**-pH
    return CO3 * (1 + h / Ks.K2 + h ** 2 / (Ks.K1 * Ks.K2))


# 8. pH and TA
def pH_TA(pH, TA, BT, PT, SiT, ST, FT, Ks):
    """
    Returns DIC

    Taken directly from MATLAB CO2SYS.
    """
    H = 10 ** -pH
    # negative alk
    BAlk = BT * Ks.KB / (Ks.KB + H)
    OH = Ks.KW / H
    PhosTop = Ks.KP1 * Ks.KP2 * H + 2 * Ks.KP1 * Ks.KP2 * Ks.KP3 - H ** 3
    PhosBot = H ** 3 + Ks.KP1 * H ** 2 + Ks.KP1 * Ks.KP2 * H + Ks.KP1 * Ks.KP2 * Ks.KP3
    PAlk = PT * PhosTop / PhosBot
    SiAlk = SiT * Ks.KSi / (Ks.KSi + H)
    # positive alk
    Hfree = H / (1 + ST / Ks.KS)
    HSO4 = ST / (1 + Ks.KS / Hfree)
    HF = FT / (1 + Ks.KF / Hfree)
    CAlk = TA - BAlk - OH - PAlk - SiAlk + Hfree + HSO4 + HF

    return CAlk * (H ** 2 + Ks.K1 * H + Ks.K1 * Ks.K2) / (Ks.K1 * (H + 2.0 * Ks.K2))


# 9. pH and DIC
def pH_DIC(pH, DIC, Ks):
    """
    Returns CO2
    """
    h = 10.0**-pH
    return DIC / (1 + Ks.K1 / h + Ks.K1 * Ks.K2 / h ** 2)


# 10. HCO3 and CO3
def HCO3_CO3(HCO3, CO3, Ks):
    """
    Returns H
    """
    HCO3, CO3 = noms(HCO3, CO3)  # get nominal values of inputs
    par = cast_array(HCO3, CO3, Ks.K1, Ks.K2)  # cast parameters into array
    shape = maxShape(HCO3, CO3, Ks.K1, Ks.K2)  # get shape of output
    
    return np.apply_along_axis(_zero_wrapper, 0, par, fn=zero_HCO3_CO3).reshape(shape)


def zero_HCO3_CO3(h, HCO3, CO3, K1, K2):
    # Roots: one pos, two neg. Use pos.
    LH = HCO3 * (h + h ** 2 / K1 + K2)
    RH = CO3 * (h + h ** 2 / K2 + h ** 3 / (K1 * K2))
    return LH - RH


# 11. HCO3 and TA
def HCO3_TA(HCO3, TA, BT, Ks):
    """
    Returns H
    """
    HCO3, TA, BT = noms(HCO3, TA, BT)  # get nominal values of inputs
    par = cast_array(
        HCO3, TA, BT, Ks.K1, Ks.K2, Ks.KB, Ks.KW
    )  # cast parameters into array
    shape = maxShape(HCO3, TA, BT, Ks.K1, Ks.K2, Ks.KB, Ks.KW)  # get shape of output

    return np.apply_along_axis(_zero_wrapper, 0, par, fn=zero_HCO3_TA).reshape(shape)


def zero_HCO3_TA(h, HCO3, TA, BT, K1, K2, KB, KW):
    # Roots: one pos, four neg. Use pos.
    LH = TA * (KB + h) * (h ** 3 + K1 * h ** 2 + K1 * K2 * h)
    RH = (
        HCO3
        * (h + h ** 2 / K1 + K2)
        * ((KB + 2 * K2) * K1 * h + 2 * KB * K1 * K2 + K1 * h ** 2)
    ) + (
        (h ** 2 + K1 * h + K1 * K2)
        * (KB * BT * h + KW * KB + KW * h - KB * h ** 2 - h ** 3)
    )
    return LH - RH


# 12. HCO3 amd DIC
def HCO3_DIC(HCO3, DIC, Ks):
    """
    Returns H
    """
    HCO3, DIC = noms(HCO3, DIC)  # get nominal values of inputs
    par = cast_array(HCO3, DIC, Ks.K1, Ks.K2)  # cast parameters into array
    shape = maxShape(HCO3, DIC, Ks.K1, Ks.K2)  # get shape of output
    
    return np.apply_along_axis(_zero_wrapper, 0, par, fn=zero_HCO3_DIC).reshape(shape)


def zero_HCO3_DIC(h, HCO3, DIC, K1, K2):
    # Roots: two pos. Use smaller.
    LH = HCO3 * (h + h ** 2 / K1 + K2)
    RH = h * DIC
    return LH - RH


# 13. CO3 and TA
def CO3_TA(CO3, TA, BT, Ks):
    """
    Returns H
    """
    CO3, TA, BT = noms(CO3, TA, BT)  # get nominal values of inputs
    par = cast_array(
        CO3, TA, BT, Ks.K1, Ks.K2, Ks.KB, Ks.KW
    )  # cast parameters into array
    shape = maxShape(CO3, TA, BT, Ks.K1, Ks.K2, Ks.KB, Ks.KW)  # get shape of output
    
    return np.apply_along_axis(_zero_wrapper, 0, par, fn=zero_CO3_TA).reshape(shape)


def zero_CO3_TA(h, CO3, TA, BT, K1, K2, KB, KW):
    # Roots: three neg, two pos. Use larger pos.
    LH = TA * (KB + h) * (h ** 3 + K1 * h ** 2 + K1 * K2 * h)
    RH = (
        CO3
        * (h + h ** 2 / K2 + h ** 3 / (K1 * K2))
        * (K1 * h ** 2 + K1 * h * (KB + 2 * K2) + 2 * KB * K1 * K2)
    ) + (
        (h ** 2 + K1 * h + K1 * K2)
        * (KB * BT * h + KW * KB + KW * h - KB * h ** 2 - h ** 3)
    )
    return LH - RH


# 14. CO3 and DIC
def CO3_DIC(CO3, DIC, Ks):
    """
    Returns H
    """
    CO3, DIC = noms(CO3, DIC)  # get nominal values of inputs
    par = cast_array(CO3, DIC, Ks.K1, Ks.K2)  # cast parameters into array
    shape = maxShape(CO3, DIC, Ks.K1, Ks.K2)  # get shape of output

    return np.apply_along_axis(_zero_wrapper, 0, par, fn=zero_CO3_DIC).reshape(shape)


def zero_CO3_DIC(h, CO3, DIC, K1, K2):
    # Roots: one pos, one neg. Use neg.
    LH = CO3 * (1 + h / K2 + h ** 2 / (K1 * K2))
    RH = DIC
    return LH - RH


# 15. TA and DIC
def TA_DIC(TA, DIC, BT, PT, SiT, ST, FT, Ks):
    """
    Returns pH

    Taken directly from MATLAB CO2SYS.
    """
    # determine shape of input
    L = maxShape(TA, DIC, BT, PT, SiT, ST, FT, Ks.K1)
    if len(L) == 0:
        L = (1,)
        
    pHguess = 8.0
    pHtol = 0.00000001
    pHx = np.full(L, pHguess)
    deltapH = np.array(pHtol + 1, ndmin=1)
    ln10 = np.log(10)

    while np.any(abs(deltapH) > pHtol):
        H = 10 ** -pHx
        # negative
        Denom = H ** 2 + Ks.K1 * H + Ks.K1 * Ks.K2
        CAlk = DIC * Ks.K1 * (H + 2 * Ks.K2) / Denom
        BAlk = BT * Ks.KB / (Ks.KB + H)
        OH = Ks.KW / H
        PhosTop = Ks.KP1 * Ks.KP2 * H + 2 * Ks.KP1 * Ks.KP2 * Ks.KP3 - H ** 3
        PhosBot = (
            H ** 3 + Ks.KP1 * H ** 2 + Ks.KP1 * Ks.KP2 * H + Ks.KP1 * Ks.KP2 * Ks.KP3
        )
        PAlk = PT * PhosTop / PhosBot
        SiAlk = SiT * Ks.KSi / (Ks.KSi + H)
        # positive
        Hfree = H / (1 + ST / Ks.KS)
        HSO4 = ST / (1 + Ks.KS / Hfree)
        HF = FT / (1 + Ks.KF / Hfree)

        Residual = TA - CAlk - BAlk - OH - PAlk - SiAlk + Hfree + HSO4 + HF

        Slope = ln10 * (
            DIC * Ks.K1 * H * (H ** 2 + Ks.K1 * Ks.K2 + 4 * H * Ks.K2) / Denom / Denom
            + BAlk * H / (Ks.KB + H)
            + OH
            + H
        )
        deltapH = Residual / Slope

        while np.any(abs(deltapH) > 1):
            FF = abs(deltapH) > 1
            deltapH[FF] = deltapH[FF] / 2

        pHx += deltapH

    return pHx


def zero_TA_DIC(h, TA, DIC, BT, K1, K2, KB, KW):
    # Roots: one pos, four neg. Use pos.
    LH = DIC * (KB + h) * (K1 * h ** 2 + 2 * K1 * K2 * h)
    RH = (TA * (KB + h) * h - KB * BT * h - KW * (KB + h) + (KB + h) * h ** 2) * (
        h ** 2 + K1 * h + K1 * K2
    )
    return LH - RH


# 1.1.9
def cCO2(H, DIC, Ks):
    """
    Returns CO2
    """
    return DIC / (1 + Ks.K1 / H + Ks.K1 * Ks.K2 / H ** 2)


# 1.1.10
def cHCO3(H, DIC, Ks):
    """
    Returns HCO3
    """
    return DIC / (1 + H / Ks.K1 + Ks.K2 / H)


# 1.1.11
def cCO3(H, DIC, Ks):
    """
    Returns CO3
    """
    return DIC / (1 + H / Ks.K2 + H ** 2 / (Ks.K1 * Ks.K2))


# 1.5.80
def cTA(H, DIC, BT, PT, SiT, ST, FT, Ks, mode="multi"):
    """
    Calculate Alkalinity. H is on Total scale.

    Returns
    -------
    If mode == 'multi' returns TA, CAlk, PAlk, SiAlk, OH
    else: returns TA
    """
    # negative
    Denom = H ** 2 + Ks.K1 * H + Ks.K1 * Ks.K2
    CAlk = DIC * Ks.K1 * (H + 2 * Ks.K2) / Denom
    BAlk = BT * Ks.KB / (Ks.KB + H)
    OH = Ks.KW / H
    PhosTop = Ks.KP1 * Ks.KP2 * H + 2 * Ks.KP1 * Ks.KP2 * Ks.KP3 - H ** 3
    PhosBot = H ** 3 + Ks.KP1 * H ** 2 + Ks.KP1 * Ks.KP2 * H + Ks.KP1 * Ks.KP2 * Ks.KP3
    PAlk = PT * PhosTop / PhosBot
    SiAlk = SiT * Ks.KSi / (Ks.KSi + H)
    # positive
    Hfree = H / (1 + ST / Ks.KS)
    HSO4 = ST / (1 + Ks.KS / Hfree)
    HF = FT / (1 + Ks.KF / Hfree)

    TA = CAlk + BAlk + OH + PAlk + SiAlk - Hfree - HSO4 - HF

    if mode == "multi":
        return TA, CAlk, BAlk, PAlk, SiAlk, OH, Hfree, HSO4, HF
    else:
        return TA

# C.4.14
def fCO2_to_CO2(fCO2, Ks):
    """
    Calculate CO2 from fCO2
    """
    return fCO2 * Ks.K0


# C.4.14
def CO2_to_fCO2(CO2, Ks):
    """
    Calculate fCO2 from CO2
    """
    return CO2 / Ks.K0


def pCO2_to_fCO2(pCO2, Tc):
    """
    Calculate fCO2 from pCO2

    Taken from matlab CO2SYS.

    This assumes that the pressure is at one atmosphere, or close to it.
    Otherwise, the Pres term in the exponent affects the results.
    Weiss, R. F., Marine Chemistry 2:203-215, 1974.

    For a mixture of CO2 and air at 1 atm (at low CO2 concentrations)
    Delta and B in cm3/mol
    """
    Tk = Tc + 273.15
    P = 1.01325  # in bar
    RT = 83.1451 * Tk

    a0, a1, a2, a3 = (-1636.75, 12.0408, -3.27957e-2, 3.16528e-05)
    b0, b1 = (57.7, -0.118)

    B = a0 + a1 * Tk + a2 * Tk ** 2 + a3 * Tk ** 3
    delta = b0 + b1 * Tk

    return pCO2 * np.exp(P * (B + 2 * delta) / RT)


def fCO2_to_pCO2(fCO2, Tc):
    """
    Calculate pCO2 from fCO2

    Taken from matlab CO2SYS.

    This assumes that the pressure is at one atmosphere, or close to it.
    Otherwise, the Pres term in the exponent affects the results.
    Weiss, R. F., Marine Chemistry 2:203-215, 1974.

    For a mixture of CO2 and air at 1 atm (at low CO2 concentrations)
    Delta and B in cm3/mol
    """
    Tk = Tc + 273.15
    P = 1.01325  # in bar
    RT = 83.1451 * Tk

    a0, a1, a2, a3 = (-1636.75, 12.0408, -3.27957e-2, 3.16528e-05)
    b0, b1 = (57.7, -0.118)

    B = a0 + a1 * Tk + a2 * Tk ** 2 + a3 * Tk ** 3
    delta = b0 + b1 * Tk

    return fCO2 / np.exp(P * (B + 2 * delta) / RT)


def calc_C_species(
    pHtot=None,
    DIC=None,
    CO2=None,
    HCO3=None,
    CO3=None,
    TA=None,
    fCO2=None,
    pCO2=None,
    T_in=None,
    S_in=None,
    BT=None,
    PT=0,
    SiT=0,
    ST=0,
    FT=0,
    Ks=None,
    **kwargs
):
    """
    Calculate all carbon species from minimal input.
    """

    # if fCO2 is given but CO2 is not, calculate CO2
    if CO2 is None:
        if fCO2 is not None:
            CO2 = fCO2_to_CO2(fCO2, Ks)
        elif pCO2 is not None:
            CO2 = fCO2_to_CO2(pCO2_to_fCO2(pCO2, T_in), Ks)

    # Carbon System Calculations (logic from Zeebe & Wolf-Gladrow, Appendix B)
    # 1. CO2 and pH
    if CO2 is not None and pHtot is not None:
        H = 10.0**-pHtot
        DIC = CO2_pH(CO2, pHtot, Ks)
    # 2. CO2 and HCO3
    elif CO2 is not None and HCO3 is not None:
        H = CO2_HCO3(CO2, HCO3, Ks)
        DIC = CO2_pH(CO2, -np.log10(H), Ks)
    # 3. CO2 and CO3
    elif CO2 is not None and CO3 is not None:
        H = CO2_CO3(CO2, CO3, Ks)
        DIC = CO2_pH(CO2, -np.log10(H), Ks)
    # 4. CO2 and TA
    elif CO2 is not None and TA is not None:
        # unit conversion because OH and H wrapped
        # up in TA fns - all need to be in same units.
        pHtot = CO2_TA(CO2=CO2, TA=TA, BT=BT, PT=PT, SiT=SiT, ST=ST, FT=FT, Ks=Ks)
        H = 10.0**-pHtot
        DIC = CO2_pH(CO2, pHtot, Ks)
    # 5. CO2 and DIC
    elif CO2 is not None and DIC is not None:
        H = CO2_DIC(CO2, DIC, Ks)
    # 6. pHtot and HCO3
    elif pHtot is not None and HCO3 is not None:
        H = 10.0**-pHtot
        DIC = pH_HCO3(pHtot, HCO3, Ks)
    # 7. pHtot and CO3
    elif pHtot is not None and CO3 is not None:
        H = 10.0**-pHtot
        DIC = pH_CO3(pHtot, CO3, Ks)
    # 8. pHtot and TA
    elif pHtot is not None and TA is not None:
        H = 10.0**-pHtot
        DIC = pH_TA(pH=pHtot, TA=TA, BT=BT, PT=PT, SiT=SiT, ST=ST, FT=FT, Ks=Ks)
    # 9. pHtot and DIC
    elif pHtot is not None and DIC is not None:
        H = 10.0**-pHtot
    # 10. HCO3 and CO3
    elif HCO3 is not None and CO3 is not None:
        H = HCO3_CO3(HCO3, CO3, Ks)
        DIC = pH_CO3(-np.log10(H), CO3, Ks)
    # 11. HCO3 and TA
    elif HCO3 is not None and TA is not None:
        Warning(
            "Nutrient alkalinity not implemented for this input combination.\nCalculations use only C and B alkalinity."
        )
        H = HCO3_TA(HCO3, TA, BT, Ks)
        DIC = pH_HCO3(-np.log10(H), HCO3, Ks)
    # 12. HCO3 amd DIC
    elif HCO3 is not None and DIC is not None:
        H = HCO3_DIC(HCO3, DIC, Ks)
    # 13. CO3 and TA
    elif CO3 is not None and TA is not None:
        Warning(
            "Nutrient alkalinity not implemented for this input combination.\nCalculations use only C and B alkalinity."
        )
        H = CO3_TA(CO3, TA, BT, Ks)
        DIC = pH_CO3(-np.log10(H), CO3, Ks)
    # 14. CO3 and DIC
    elif CO3 is not None and DIC is not None:
        H = CO3_DIC(CO3, DIC, Ks)
    # 15. TA and DIC
    elif TA is not None and DIC is not None:
        pHtot = TA_DIC(TA=TA, DIC=DIC, BT=BT, PT=PT, SiT=SiT, ST=ST, FT=FT, Ks=Ks)
        H = 10.0**-pHtot

    # The above makes sure that DIC and H are known,
    # this next bit calculates all the missing species
    # from DIC and H.
    if CO2 is None:
        CO2 = cCO2(H, DIC, Ks)
    if fCO2 is None:
        fCO2 = CO2_to_fCO2(CO2, Ks)
    if pCO2 is None:
        pCO2 = fCO2_to_pCO2(fCO2, T_in)
    if HCO3 is None:
        HCO3 = cHCO3(H, DIC, Ks)
    if CO3 is None:
        CO3 = cCO3(H, DIC, Ks)
    # Calculate all elements of Alkalinity
    (TA, CAlk, BAlk, PAlk, SiAlk, OH, Hfree, HSO4, HF) = cTA(
        H=H, DIC=DIC, BT=BT, PT=PT, SiT=SiT, ST=ST, FT=FT, Ks=Ks, mode="multi"
    )

    # if pH not calced yet, calculate on all scales.
    if pHtot is None:
        pHtot = np.array(-np.log10(H), ndmin=1)
    
    FREEtoTOT = -np.log10((1 + ST / Ks.KS))
    SWStoTOT = -np.log10((1 + ST / Ks.KS) / (1 + ST / Ks.KS + FT / Ks.KF))
    fH = calc_fH(T_in + 273.15, S_in)
    
    return Bunch(
        {
            "pHtot": pHtot,
            "pHfree": pHtot - FREEtoTOT,
            "pHsws": pHtot - SWStoTOT,
            "pHNBS": pHtot - SWStoTOT - np.log10(fH),
            "TA": TA,
            "DIC": DIC,
            "CO2": CO2,
            "H": H,
            "HCO3": HCO3,
            "fCO2": fCO2,
            "pCO2": pCO2,
            "CO3": CO3,
            "CAlk": CAlk,
            "BAlk": BAlk,
            "PAlk": PAlk,
            "SiAlk": SiAlk,
            "OH": OH,
            "Hfree": Hfree,
            "HSO4": HSO4,
            "HF": HF,
        }
    )


def calc_revelle_factor(TA, DIC, BT, PT, SiT, ST, FT, Ks):
    """
    Calculate Revelle Factor

    (dpCO2 / dDIC)
    """
    dDIC = 1e-6  # (1 umol kg-1)

    pH = TA_DIC(TA=TA, DIC=DIC, BT=BT, PT=PT, SiT=SiT, ST=ST, FT=FT, Ks=Ks)
    fCO2 = cCO2(10.0**-pH, DIC, Ks) / Ks.K0

    # Calculate new fCO2 above and below given value
    pH_hi = TA_DIC(TA=TA, DIC=DIC + dDIC, BT=BT, PT=PT, SiT=SiT, ST=ST, FT=FT, Ks=Ks)
    fCO2_hi = cCO2(10.0**-pH_hi, DIC, Ks) / Ks.K0

    pH_lo = TA_DIC(TA=TA, DIC=DIC - dDIC, BT=BT, PT=PT, SiT=SiT, ST=ST, FT=FT, Ks=Ks)
    fCO2_lo = cCO2(10.0**-pH_lo, DIC, Ks) / Ks.K0

    return (fCO2_hi - fCO2_lo) * DIC / (fCO2 * 2 * dDIC)



"""
Functions for calculating the carbon and boron chemistry of seawater.
"""

import numpy as np
import pandas as pd
from cbsyst.carbon import calc_C_species, calc_revelle_factor, pCO2_to_fCO2, fCO2_to_CO2
from cbsyst.boron import calc_B_species
from cbsyst.boron_isotopes import d11_to_A11, A11_to_d11, get_alphaB, calc_B_isotopes
from cbsyst.helpers import Bunch, NnotNone, calc_FT, calc_ST, calc_BT, calc_pH_scales
from kgen import calc_Ks

# C Speciation
# ------------
def Csys(
    pHtot=None, DIC=None, TA=None,
    CO2=None, HCO3=None, CO3=None,
    pCO2=None, fCO2=None,
    BT=None,
    Ca=0.0102821, Mg=0.0528171,
    T_in=25.0, T_out=None, 
    S_in=35.0, S_out=None,
    P_in=0.0, P_out=None,
    PT=0.0, SiT=0.0,
    ST=None, FT=None,
    pHsws=None, pHfree=None, pHNBS=None,
    unit="umol", Ks=None,
    pdict=None,
    OmegaC=None, OmegaA=None,
    MyAMI_mode="calculate"
):
    """
    Calculate the carbon chemistry of seawater from a minimal parameter set.

    Constants calculated by MyAMI model (Hain et al, 2015; doi:10.1002/2014GB004986).
    Speciation calculations from Zeebe & Wolf-Gladrow (2001; ISBN:9780444509468) Appendix B

    Inputs must either be single values, arrays of equal length or a mixture of both.
    If you use arrays of unequal length, it won't work.

    Error propagation:
    If inputs are ufloat or uarray (from uncertainties package) errors will
    be propagated through all calculations, but:

    **WARNING** Error propagation NOT IMPLEMENTED for carbon system calculations
    with zero-finders (i.e. when pH is not given; cases 2-5 and 10-15).

    Concentration Units
    +++++++++++++++++++
    * Ca and Mg must be in molar units.
    * All other units must be the same, and can be specified in the 'unit' variable. Defaults to umolar.

    Parameters
    ----------
    pH, DIC, CO2, HCO3, CO3, TA, OmegaC, OmegaA : array-like
        Carbon system parameters. Two of these must be provided.
    BT : array-like
        Total B at the input salinity (in umol/kg) used in Alkalinity calculations. 
        If missing, this is calculated as 0.000416 * Sal/35 
        (Uppstrom et al. 1974).
    Ca, Mg : array-like
        The [Ca] and [Mg] of STANDARD seawater (i.e. 35 salinity), in mol / kg.
        Used in calculating MyAMI constants.
    T_in, S_in : array-like
        Temperature in Celcius and Salinity in PSU that the
        measurements were conducted under.
        Used in calculating constants.
    P_in : array-like
        Pressure in Bar that the measurements were conducted under.
        Used in pressure-correcting constants.
    T_out, S_out : array-like
        Temperature in Celcius and Salinity in PSU of the desired
        output conditions.
        Used in calculating constants.
    P_in : array-like
        Pressure in Bar of the desired output conditions.
        Used in pressure-correcting constants.
    unit : str
        Concentration units of C and B parameters (all must be in
        the same units).
        Can be 'mol', 'mmol', 'umol', 'nmol', 'pmol' or 'fmol'.
        Used in calculating Alkalinity. Default is 'umol'.
    Ks : dict
        A dictionary of constants. Must contain keys
        'K1', 'K2', 'KB' and 'KW'.
        If None, Ks are calculated using MyAMI model.
    pdict : dict
        Optionally, you can provide some or all parameters as a dict,
        with keys the same as the parameter names above. Any parameters
        included in the dict will overwrite manually specified
        parameters. This is particularly useful if you're including
        this in other code.

    Returns
    -------
    dict(/Bunch) containing all calculated parameters.
    """

    # Bunch inputs
    ps = Bunch(locals())
    if isinstance(pdict, dict):
        ps.update(pdict)

    # convert unit to multiplier
    udict = {
        "mol": 1.0,
        "mmol": 1.0e3,
        "umol": 1.0e6,
        "nmol": 1.0e9,
        "pmol": 1.0e12,
        "fmol": 1.0e15,
    }
    if isinstance(ps.unit, str):
        ps.unit = udict[ps.unit]

    upar = ["DIC", "TA", "CO2", "HCO3", "CO3", "BT", "PT", "SiT"]
    if ps.unit != 1:    
        for p in upar:
            if ps[p] is not None:
                ps[p] = np.divide(ps[p], ps.unit)  # convert to molar
    
    gupar = ['pCO2', 'fCO2']  # special unit case: pCO2, fCO2 because pCO2 is always in ppm
    for p in gupar:
        if ps[p] is not None:
            ps[p] = np.divide(ps[p], 1e6)  # convert to mole fraction
    
    # Conserved seawater chemistry
    if ps.ST is None:
        ps.ST = calc_ST(ps.S_in)
    if ps.FT is None:
        ps.FT = calc_FT(ps.S_in)
    if ps.BT is None:
        ps.BT = calc_BT(ps.S_in)

    # Remove negative values 
    for p in ["DIC", "CO2", "HCO3", "CO3", "BT", "fCO2", "pCO2", "PT", "SiT"]:
        if ps[p] is not None:
            if isinstance(ps[p], (np.ndarray, pd.core.series.Series)):
                if np.any(ps[p] < 0):
                    ps[p][ps[p] < 0] = np.nan
            elif ps[p] < 0:
                ps[p] = np.nan

    
    # Calculate Ks at input conditions
    if isinstance(Ks, dict):
        ps.Ks = Bunch(Ks)
    else:
        ps.Ks = Bunch(calc_Ks(temp_c=ps.T_in, sal=ps.S_in, p_bar=ps.P_in, magnesium=ps.Mg, calcium=ps.Ca, sulphate=ps.ST, fluorine=ps.FT, MyAMI_mode=ps.MyAMI_mode))
        
    # Calculate pH scales at input conditions (does nothing if no pH given)
    ps.update(
        calc_pH_scales(
            pHtot=ps.pHtot,
            pHfree=ps.pHfree,
            pHsws=ps.pHsws,
            pHNBS=ps.pHNBS,
            ST=ps.ST,
            FT=ps.FT,
            TempK=ps.T_in + 273.15,
            Sal=ps.S_in,
            Ks=ps.Ks
        )
    )
    
    # If OmegaA or OmegaC are provided, calculate CO3 from them
    if OmegaA is not None:
        ps['CO3'] = OmegaA * ps.Ks.KspA / (ps.Ca * ps.S_in / 35.)
    elif OmegaC is not None:
        ps['CO3'] = OmegaC * ps.Ks.KspC / (ps.Ca * ps.S_in / 35.)
    
    # calculate C system at input conditions
    ps.update(calc_C_species(**ps))
    
    ps["revelle_factor"] = calc_revelle_factor(
        TA=ps.TA,
        DIC=ps.DIC,
        BT=ps.BT,
        PT=ps.PT,
        SiT=ps.SiT,
        ST=ps.ST,
        FT=ps.FT,
        Ks=ps.Ks,
    )
    
    # calc Omega
    oCa = ps.Ca * ps.S_in / 35.
    ps['OmegaA'] = ps['CO3'] * oCa / ps.Ks.KspA
    ps['OmegaC'] = ps['CO3'] * oCa / ps.Ks.KspC

    # clean up output
    outputs = [
        "BT", "CO2", "CO3", "Ca", "DIC", "H", "HCO3", 
        "Mg", "S_in", "T_in", "TA", "CAlk", "PAlk", 
        "SiAlk", "OH", 'OmegaA', 'OmegaC', 'revelle_factor']
    for k in outputs:
        if not isinstance(ps[k], np.ndarray):
            # convert all outputs to (min) 1D numpy arrays.
            ps[k] = np.array(ps[k], ndmin=1)
    if ps.unit != 1:
        for p in upar + ["CAlk", "BAlk", "PAlk", "SiAlk", "OH", "HSO4", "HF", "Hfree"]:
            ps[p] *= ps.unit  # convert back to input units
        for p in gupar:
            ps[p] = np.multiply(ps[p], 1e6)  # convert to ppm

    # Calculate Output Conditions
    # ===========================
    if ps.T_out is not None or ps.S_out is not None or ps.P_out is not None:
        if ps.T_out is None:
            ps.T_out = ps.T_in
        if ps.S_out is None:
            ps.S_out = ps.S_in
        else:
            # if salinity is modified, update salinity-dependent parameters
            BT = ps.BT * ps.S_out / ps.S_in
            ST = ps.ST * ps.S_out / ps.S_in
            FT = ps.FT * ps.S_out / ps.S_in
        if ps.P_out is None:
            ps.P_out = ps.P_in
        # assumes conserved alkalinity and DIC
        
        # This needs to be different depending on whether T or S is changing.
        out_cond = Csys(
            TA=ps.TA,
            DIC=ps.DIC,
            T_in=ps.T_out,
            S_in=ps.S_out,
            P_in=ps.P_out,
            unit=ps.unit,
            Ca=Ca,
            Mg=Mg,
            BT=BT,
            FT=FT,
            ST=ST,
        )

        # rename parameters in output conditions
        outputs = [
            "BAlk", "BT", "CAlk", "CO2", "CO3", "DIC", "H", "HCO3", 
            "HF", "HSO4", "Hfree", "Ks", "OH", "PAlk", "SiAlk", "TA", "FT",
            "PT", "ST", "SiT", "fCO2", "pCO2", "pHfree", "pHsws", 
            "pHtot", "pHNBS", 'OmegaA', 'OmegaC', "revelle_factor",
        ]

        ps.update({k + "_in": ps[k] for k in outputs})
        ps.update({k: out_cond[k] for k in outputs})

    # remove some superfluous outputs
    rem = ["pdict"]
    for r in rem:
        if r in ps:
            del ps[r]

    return ps


# B Speciation
# ------------
def Bsys(
    pHtot=None,
    BT=None,
    BO3=None,
    BO4=None,
    ABT=None,
    ABO3=None,
    ABO4=None,
    dBT=None,
    dBO3=None,
    dBO4=None,
    alphaB=None,
    T_in=25.0,
    S_in=35.0,
    P_in=0.0,
    Ca=0.0102821,
    Mg=0.0528171,
    ST=None,
    FT=None,
    pHsws=None,
    pHfree=None,
    pHNBS=None,
    Ks=None,
    pdict=None,
    MyAMI_mode="calculate"
):
    """
    Calculate the boron chemistry of seawater from a minimal parameter set.

    Constants calculated by MyAMI model (Hain et al, 2015; doi:10.1002/2014GB004986).
    Speciation calculations from Zeebe & Wolf-Gladrow (2001; ISBN:9780444509468).

    Inputs must either be single values, arrays of equal length or a mixture of both.
    If you use arrays of unequal length, it won't work.

    Error propagation:
    If inputs are ufloat or uarray (from uncertainties package) errors will
    be propagated through all calculations.

    Concentration Units
    +++++++++++++++++++
    * All concentrations must be in the same units. Returned in the same units as inputs.

    Parameters
    ----------
    pH, BT, BO3, BO4 : array-like
        Boron system parameters. Two of these must be provided.
    dBT, dBO3, dBO4, ABT, ABO3, ABO4 : array-like
        delta (d) or fractional abundance (A) values for the Boron
        isotope system. One of these must be provided.
    alphaB : array-like
        The alpha value for BO3-BO4 isotope fractionation.
    T, S : array-like
        Temperature in Celcius and Salinity in PSU.
        Used in calculating MyAMI constants.
    P : array-like
        Pressure in Bar.
        Used in calculating MyAMI constants.
    Ca, Mg : arra-like
        The [Ca] and [Mg] of STANDARD seawater (i.e. 35 salinity), in mol / kg.
        Used in calculating MyAMI constants.
    Ks : dict
        A dictionary of constants. Must contain keys
        'K1', 'K2', 'KB' and 'KW'.
        If None, Ks are calculated using MyAMI model.
    pdict : dict
        Optionally, you can provide some or all parameters as a dict,
        with keys the same as the parameter names above. Any parameters
        included in the dict will overwrite manually specified
        parameters. This is particularly useful if you're including
        this in other code.

    Returns
    -------
    dict(/Bunch) containing all calculated parameters.
    """
    # input checks
    if NnotNone(BT, BO3, BO4) < 1:
        raise ValueError("Must provide at least one of BT, BO3 or BO4")
    if NnotNone(dBT, dBO3, dBO4, ABT, ABO3, ABO4) < 1:
        raise ValueError("Must provide one of dBT, dBO3, dBO4, ABT, ABO3 or ABO4")

    # Bunch inputs
    ps = Bunch(locals())
    if isinstance(pdict, dict):
        ps.update(pdict)

    # Conserved seawater chemistry
    if ps.ST is None:
        ps.ST = calc_ST(ps.S_in)
    if ps.FT is None:
        ps.FT = calc_FT(ps.S_in)
    
    # Remove negative values 
    for p in ["BT", "BO3", "BO4", "ST", "FT"]:
        if ps[p] is not None:
            if isinstance(ps[p], (np.ndarray, pd.core.series.Series)):
                ps[p][ps[p] < 0] = np.nan
            elif ps[p] < 0:
                ps[p] = np.nan

    # Calculate Ks
    ps.Ks = Bunch(calc_Ks(temp_c=ps.T_in, sal=ps.S_in, p_bar=ps.P_in, magnesium=ps.Mg, calcium=ps.Ca, sulphate=ps.ST, fluorine=ps.FT, MyAMI_mode=ps.MyAMI_mode))

    # Calculate pH scales (does nothing if no pH given)
    ps.update(
        calc_pH_scales(
            pHtot=ps.pHtot,
            pHfree=ps.pHfree,
            pHsws=ps.pHsws,
            pHNBS=ps.pHNBS,
            ST=ps.ST,
            FT=ps.FT,
            TempK=ps.T_in + 273.15,
            Sal=ps.S_in,
            Ks=ps.Ks,
        )
    )
    
    # calcualte pH if not provided.
    if ps.pHtot is None:
        if ps.dBT is None and ps.ABT is None:
            ps.dBT = 39.61
        if ps.dBT is not None:
            ps.ABT = d11_to_A11(ps.dBT)
        if ps.dBO3 is not None:
            ps.ABO3 = d11_to_A11(ps.dBO3)
        if ps.dBO4 is not None:
            ps.ABO4 = d11_to_A11(ps.dBO4)
        if ps.alphaB is None:
            ps.alphaB = get_alphaB()
            
        ps.update(calc_B_isotopes(**ps))

    ps.update(calc_B_species(**ps))

    # If pH not calced yet, calculate on all scales (does nothing if all pH scales already calculated)
    ps.update(
        calc_pH_scales(
            pHtot=ps.pHtot,
            pHfree=ps.pHfree,
            pHsws=ps.pHsws,
            pHNBS=ps.pHNBS,
            ST=ps.ST,
            FT=ps.FT,
            TempK=ps.T_in + 273.15,
            Sal=ps.S_in,
            Ks=ps.Ks,
        )
    )

    # If any isotope parameter specified, calculate the isotope systen.
    if NnotNone(ps.ABT, ps.ABO3, ps.ABO4, ps.dBT, ps.dBO3, ps.dBO4) != 0:
        ps.update(ABsys(pdict=ps))

    for k in ["BT", "H", "BO3", "BO4", "Ca", "Mg", "S_in", "T_in"]:
        # convert all outputs to (min) 1D numpy arrays.
        if not isinstance(ps[k], np.ndarray):
            # convert all outputs to (min) 1D numpy arrays.
            ps[k] = np.array(ps[k], ndmin=1)

    # remove some superfluous outputs
    rem = ["pdict"]
    for r in rem:
        if r in ps:
            del ps[r]

    return ps


# B Isotopes
# ----------
def ABsys(
    pHtot=None,
    ABT=None,
    ABO3=None,
    ABO4=None,
    dBT=None,
    dBO3=None,
    dBO4=None,
    alphaB=None,
    T_in=25.0,
    S_in=35.0,
    P_in=0.0,
    Ca=0.0102821,
    Mg=0.0528171,
    ST=None,
    FT=None,
    pHsws=None,
    pHfree=None,
    pHNBS=None,
    Ks=None,
    pdict=None,
    MyAMI_mode="calculate"
):
    """
    Calculate the boron isotope chemistry of seawater from a minimal parameter set.

    Constants calculated by MyAMI model (Hain et al, 2015; doi:10.1002/2014GB004986).
    Speciation calculations from Zeebe & Wolf-Gladrow (2001; ISBN:9780444509468).

    Inputs must either be single values, arrays of equal length or a mixture of both.
    If you use arrays of unequal length, it won't work.

    Error propagation:
    If inputs are ufloat or uarray (from uncertainties package) errors will
    be propagated through all calculations.

    Concentration Units
    +++++++++++++++++++
    * 'A' is fractional abundance (11B / BT)
    * 'd' are delta values
    Either specified, both returned.

    Parameters
    ----------
    pH, ABT, ABO3, ABO4, dBT, dBO3, dBO4 : array-like
        Boron isotope system parameters. Two of pH, {ABT, dBT},
        {ABO4, dBO4}, {ABO3, dBO3} must be provided.
    alphaB : array-like
        Alpha value describing B fractionation (1.0XXX).
        If missing, it's calculated using the temperature
        sensitive formulation of Honisch et al (2008)
    T, S : array-like
        Temperature in Celcius and Salinity in PSU.
        Used in calculating MyAMI constants.
    P : array-like
        Pressure in Bar.
        Used in calculating MyAMI constants.
    Ca, Mg : arra-like
        The [Ca] and [Mg] of STANDARD seawater (i.e. 35 salinity), in mol / kg.
        Used in calculating MyAMI constants.
    Ks : dict
        A dictionary of constants. Must contain keys
        'K1', 'K2', 'KB' and 'KW'.
        If None, Ks are calculated using MyAMI model.
    pdict : dict
        Optionally, you can provide some or all parameters as a dict,
        with keys the same as the parameter names above. Any parameters
        included in the dict will overwrite manually specified
        parameters. This is particularly useful if you're including
        this in other code.

    Returns
    -------
    dict(/Bunch) containing all calculated parameters.
    """

    # Bunch inputs
    ps = Bunch(locals())
    if isinstance(pdict, dict):
        ps.update(pdict)

    # Conserved seawater chemistry
    if ps.ST is None:
        ps.ST = calc_ST(ps.S_in)
    if ps.FT is None:
        ps.FT = calc_FT(ps.S_in)

    # Calculate Ks
    if isinstance(Ks, dict):
        ps.Ks = Bunch(Ks)
    else:
        ps.Ks = Bunch(calc_Ks(temp_c=ps.T_in, sal=ps.S_in, p_bar=ps.P_in, magnesium=ps.Mg, calcium=ps.Ca, sulphate=ps.ST, fluorine=ps.FT, MyAMI_mode=ps.MyAMI_mode))

    # Calculate pH scales (does nothing if no pH given)
    ps.update(
        calc_pH_scales(
            pHtot=ps.pHtot,
            pHfree=ps.pHfree,
            pHsws=ps.pHsws,
            pHNBS=ps.pHNBS,
            ST=ps.ST,
            FT=ps.FT,
            TempK=ps.T_in + 273.15,
            Sal=ps.S_in,
            Ks=ps.Ks,
        )
    )

    # if deltas provided, calculate corresponding As
    if ps.dBT is not None:
        ps.ABT = d11_to_A11(ps.dBT)
    if ps.dBO3 is not None:
        ps.ABO3 = d11_to_A11(ps.dBO3)
    if ps.dBO4 is not None:
        ps.ABO4 = d11_to_A11(ps.dBO4)

    # calculate alpha
    if alphaB is None:
        ps.alphaB = get_alphaB()
    else:
        ps.alphaB = alphaB

    ps.update(calc_B_isotopes(**ps))

    if ps.dBT is None:
        ps.dBT = A11_to_d11(ps.ABT)
    if ps.dBO3 is None:
        ps.dBO3 = A11_to_d11(ps.ABO3)
    if ps.dBO4 is None:
        ps.dBO4 = A11_to_d11(ps.ABO4)

    for k in [
        "ABO3",
        "ABO4",
        "ABT",
        "Ca",
        "H",
        "Mg",
        "S_in",
        "T_in",
        "alphaB",
        "dBO3",
        "dBO4",
        "dBT",
        "pHtot",
    ]:
        if not isinstance(ps[k], np.ndarray):
            # convert all outputs to (min) 1D numpy arrays.
            ps[k] = np.array(ps[k], ndmin=1)

    # remove some superfluous outputs
    rem = ["pdict"]
    for r in rem:
        if r in ps:
            del ps[r]

    return ps


# Whole C-B-Isotope System
# ------------------------
def CBsys(
    pHtot=None,
    DIC=None,
    CO2=None,
    HCO3=None,
    CO3=None,
    TA=None,
    fCO2=None,
    pCO2=None,
    BT=None,
    BO3=None,
    BO4=None,
    ABT=None,
    ABO3=None,
    ABO4=None,
    dBT=None,
    dBO3=None,
    dBO4=None,
    alphaB=None,
    T_in=25.0,
    S_in=35.0,
    P_in=0.0,
    T_out=None,
    S_out=None,
    P_out=None,
    Ca=0.0102821,
    Mg=0.0528171,
    PT=0.0,
    SiT=0.0,
    ST=None,
    FT=None,
    pHsws=None,
    pHfree=None,
    pHNBS=None,
    Ks=None,
    pdict=None,
    unit="umol",
    MyAMI_mode="calculate"
):
    """
    Calculate carbon, boron and boron isotope chemistry of seawater from a minimal parameter set.

    Constants calculated by MyAMI model (Hain et al, 2015; doi:10.1002/2014GB004986).
    Speciation calculations from Zeebe & Wolf-Gladrow (2001; ISBN:9780444509468) Appendix B

    Inputs must either be single values, arrays of equal length or a mixture of both.
    If you use arrays of unequal length, it won't work.

    Note: Special Case! If pH is not known, you must provide either:
      - Two of [DIC, CO2, HCO3, CO3], and one of [BT, BO3, BO4]
      - One of [DIC, CO2, HCO3, CO3], and TA and BT
      - Two of [BT, BO3, BO4] and one of [DIC, CO2, HCO3, CO3]

    Isotopes will only be calculated if one of [ABT, ABO3, ABO4, dBT, dBO3, dBO4]
    is provided.

    Error propagation:
    If inputs are ufloat or uarray (from uncertainties package) errors will
    be propagated through all calculations, but:

    **WARNING** Error propagation NOT IMPLEMENTED for carbon system calculations
    with zero-finders (i.e. when pH is not given; cases 2-5 and 10-15).

    Concentration Units
    +++++++++++++++++++
    * Ca and Mg must be in molar units.
    * All other units must be the same, and can be specified in the 'unit' variable. Defaults to umolar.
    * Isotopes can be in A (11B / BT) or d (delta). Either specified, both returned.

    Parameters
    ----------
    pH, DIC, CO2, HCO3, CO3, TA : array-like
        Carbon system parameters. Two of these must be provided.
        If TA is specified, a B species must also be specified.
    pH, BT, BO3, BO4 : array-like
        Boron system parameters. Two of these must be provided.
    pH, ABT, ABO3, ABO4, dBT, dBO3, dBO4 : array-like
        Boron isotope system parameters. pH and one other
        parameter must be provided.
    alphaB : array-like
        Alpha value describing B fractionation (1.0XXX).
        If missing, it's calculated using the temperature
        sensitive formulation of Honisch et al (2008)
    T, S : array-like
        Temperature in Celcius and Salinity in PSU.
        Used in calculating MyAMI constants.
    P : array-like
        Pressure in Bar.
        Used in calculating MyAMI constants.
    unit : str
        Concentration units of C and B parameters (all must be in
        the same units).
        Can be 'mol', 'mmol', 'umol', 'nmol', 'pmol' or 'fmol'.
        Used in calculating Alkalinity. Default is 'umol'.
    Ca, Mg : arra-like
        The [Ca] and [Mg] of STANDARD seawater (i.e. 35 salinity), in mol / kg.
        Used in calculating MyAMI constants.
    Ks : dict
        A dictionary of constants. Must contain keys
        'K1', 'K2', 'KB' and 'KW'.
        If None, Ks are calculated using MyAMI model.
    pdict : dict
        Optionally, you can provide some or all parameters as a dict,
        with keys the same as the parameter names above. Any parameters
        included in the dict will overwrite manually specified
        parameters. This is particularly useful if you're including
        this in other code.

    Returns
    -------
    dict(/Bunch) containing all calculated parameters.
    """
    # Bunch inputs
    ps = Bunch(locals())
    if isinstance(pdict, dict):
        ps.update(pdict)

    # convert unit to multiplier
    udict = {
        "mol": 1.0,
        "mmol": 1.0e3,
        "umol": 1.0e6,
        "nmol": 1.0e9,
        "pmol": 1.0e12,
        "fmol": 1.0e15,
    }
    if isinstance(ps.unit, str):
        ps.unit = udict[ps.unit]
    elif isinstance(ps.unit, (int, float)):
        ps.unit = unit

    upar = [
        "DIC",
        "CO2",
        "HCO3",
        "CO3",
        "TA",
        "BT",
        "BO3",
        "BO4",
        "PT",
        "SiT",
    ]
 
    for p in upar:
        if ps[p] is not None:
            ps[p] = np.divide(ps[p], ps.unit)  # convert to molar

    gupar = ['pCO2', 'fCO2']  # special unit case: pCO2, fCO2 because pCO2 is always in ppm
    for p in gupar:
        if ps[p] is not None:
            ps[p] = np.divide(ps[p], 1e6)  # convert to mole fraction
    
    # Conserved seawater chemistry
    if ps.ST is None:
        ps.ST = calc_ST(ps.S_in)
    if ps.FT is None:
        ps.FT = calc_FT(ps.S_in)

    # Remove negative values 
    for p in ["DIC", "CO2", "HCO3", "CO3", "BT", "BO3", "BO4", "fCO2", "pCO2", "PT", "SiT"]:
        if ps[p] is not None:
            if isinstance(ps[p], (np.ndarray, pd.core.series.Series)):
                ps[p][ps[p] < 0] = np.nan
            elif ps[p] < 0:
                ps[p] = np.nan
    
    # Calculate Ks
    if isinstance(Ks, dict):
        ps.Ks = Bunch(Ks)
    else:
        ps.Ks = Bunch(calc_Ks(temp_c=ps.T_in, sal=ps.S_in, p_bar=ps.P_in, magnesium=ps.Mg, calcium=ps.Ca, sulphate=ps.ST, fluorine=ps.FT, MyAMI_mode=ps.MyAMI_mode))

    # calculate alpha
    if alphaB is None:
        ps.alphaB = get_alphaB()
    else:
        ps.alphaB = alphaB
        
    # convert any B isotopes to A notation
    if ps.dBT is None and ps.ABT is None:
        ps.dBT = 39.61
    if ps.dBT is not None:
        ps.ABT = d11_to_A11(ps.dBT)
    if ps.dBO3 is not None:
        ps.ABO3 = d11_to_A11(ps.dBO3)
    if ps.dBO4 is not None:
        ps.ABO4 = d11_to_A11(ps.dBO4)
        
    nBiso = NnotNone(ps.ABT) + NnotNone(ps.ABO4, ps.ABO3)
    
    # Calculate all pH scales (does nothing if no pH given)
    ps.update(
        calc_pH_scales(
            pHtot=ps.pHtot,
            pHfree=ps.pHfree,
            pHsws=ps.pHsws,
            pHNBS=ps.pHNBS,
            ST=ps.ST,
            FT=ps.FT,
            TempK=ps.T_in + 273.15,
            Sal=ps.S_in,
            Ks=ps.Ks,
        )
    )
    
    # if fCO2 is given but CO2 is not, calculate CO2
    if ps.CO2 is None:
        if ps.fCO2 is not None:
            ps.CO2 = fCO2_to_CO2(ps.fCO2, ps.Ks)
        elif ps.pCO2 is not None:
            ps.CO2 = fCO2_to_CO2(pCO2_to_fCO2(ps.pCO2, ps.T_in), ps.Ks)
    
    # if no B info provided, assume modern conc.
    nBspec = NnotNone(ps.BT, ps.BO3, ps.BO4)
    if nBspec == 0:
        ps.BT = calc_BT(ps.S_in)

    # count number of not None C parameters
    nCspec = NnotNone(ps.DIC, ps.CO2, ps.HCO3, ps.CO3)  # used below
        
    # if pH or two B species are given:
    if ps.pHtot is not None or nBspec == 2:
        ps.update(calc_B_species(**ps))
        ps.update(calc_C_species(**ps))
        ps.update(calc_B_isotopes(**ps))
    # if pH can be calculated from B isotopes
    elif nBiso == 2:
        ps.update(calc_B_isotopes(**ps))
        ps.update(calc_B_species(**ps))
        ps.update(calc_C_species(**ps))
    # if ther eare two carbon species, or one carbon species + TA and BT
    elif (nCspec == 2) | ((nCspec == 1) & (NnotNone(ps.TA, ps.BT) == 2)):
        ps.update(calc_C_species(**ps))
        ps.update(calc_B_species(**ps))
        ps.update(calc_B_isotopes(**ps))
    
    else:  # if neither condition is met, throw an error
        raise ValueError(
            (
                "Impossible! You haven't provided enough information.\n"
                + "If you don't know pH, you must provide either:\n"
                + "  - Two of [DIC, CO2, HCO3, CO3] and BT\n"
                + "  - One of [DIC, CO2, HCO3, CO3], and TA and BT\n"
                + "  - Two of [BT, BO3, BO4] and one of [DIC, CO2, HCO3, CO3]"
                + "  - Two of [dBT, dBO3, dBO4] and one of [DIC, CO2, HCO3, CO3]"                
            )
        )

    # convert isotopes to delta notation
    if ps.dBT is None:
        ps.dBT = A11_to_d11(ps.ABT)
    if ps.dBO3 is None:
        ps.dBO3 = A11_to_d11(ps.ABO3)
    if ps.dBO4 is None:
        ps.dBO4 = A11_to_d11(ps.ABO4)
    
    ps["revelle_factor"] = calc_revelle_factor(
        TA=ps.TA,
        DIC=ps.DIC,
        BT=ps.BT,
        PT=ps.PT,
        SiT=ps.SiT,
        ST=ps.ST,
        FT=ps.FT,
        Ks=ps.Ks,
    )

    # calc Omega
    if Ca is None:
        oCa = 0.0102821 * ps.S_in / 35.
    else:
        oCa = Ca * ps.S_in / 35.
    ps['OmegaA'] = ps['CO3'] * oCa / ps.Ks.KspA
    ps['OmegaC'] = ps['CO3'] * oCa / ps.Ks.KspC
    
    # clean up output
    outputs = [
        "BAlk",
        "BT",
        "CAlk",
        "CO2",
        "CO3",
        "DIC",
        "H",
        "HCO3",
        "HF",
        "HSO4",
        "Hfree",
        "Ks",
        "OH",
        "PAlk",
        "SiAlk",
        "TA",
        "FT",
        "PT",
        "ST",
        "SiT",
        "fCO2",
        "pCO2",
        "pHfree",
        "pHsws",
        "pHtot",
        "pHNBS",
        'OmegaA',
        'OmegaC',
        'revelle_factor',
        "BO3",
        "BO4",
        "ABO3",
        "ABO4",
        "dBO3",
        "dBO4",
    ]
    for k in outputs:
        if k == 'Ks':
            continue
        if not isinstance(ps[k], np.ndarray):
            # convert all outputs to (min) 1D numpy arrays.
            ps[k] = np.array(ps[k], ndmin=1)

    # Handle Units - at this point everything is in mol/kg-SW
    for p in upar + ["CAlk", "BAlk", "PAlk", "SiAlk", "OH", "HSO4", "HF", "Hfree"]:
        ps[p] *= ps.unit  # convert back to input units

    for p in gupar:
        if ps[p] is not None:
            ps[p] = np.multiply(ps[p], 1e6)  # convert to ppm

    # Calculate Output Conditions
    # ===========================
    
    # Recursive approach to calculate output params.
    # if output conditions specified, calculate outputs.
    if ps.T_out is not None or ps.S_out is not None or ps.P_out is not None:
        if ps.T_out is None:
            ps.T_out = ps.T_in
        if ps.S_out is None:
            ps.S_out = ps.S_in
        else:
            # if salinity is modified, update salinity-dependent parameters
            BT = ps.BT * ps.S_out / ps.S_in
            ST = ps.ST * ps.S_out / ps.S_in
            FT = ps.FT * ps.S_out / ps.S_in
        if ps.P_out is None:
            ps.P_out = ps.P_in
        # assumes conserved alkalinity, DIC and BT
        out_cond = CBsys(
            TA=ps.TA,
            DIC=ps.DIC,
            dBT=ps.dBT,
            T_in=ps.T_out,
            S_in=ps.S_out,
            P_in=ps.P_out,
            unit=ps.unit,
            Ca=Ca,
            Mg=Mg,
            BT=BT,
            FT=FT,
            ST=ST,
        )
        # rename parameters in output conditions
        ps.update({k + "_in": ps[k] for k in outputs})
        ps.update({k: out_cond[k] for k in outputs})

        # remove some superfluous outputs
    rem = ["pdict", "unit"]
    for r in rem:
        if r in ps:
            del ps[r]
    return ps



class const:
    def __init__(self, T, S, P, Mg=None, Ca=None, TS=None, TF=None, TB=None, mode='calculate'):
        if TS is None:
            TS = calc_TS(S)
        if TF is None:
            TF = calc_TF(S)
        
        self.TS = TS
        self.TF = TF
        self.Mg = Mg
        self.Ca = Ca
        
        self.Ks = calc_Ks(T=T, S=S, P=P, Mg=Mg, Ca=Ca, TS=TS, TF=TF, MyAMI_Mode=mode)


import os
import pandas as pd
import numpy as np
import cbsyst as cb
import matplotlib.pyplot as plt


def GLODAPv2_comparison(figdir="."):

    print("\n********************************************")
    print("Generating GLODAPv2 Comparison Plots")
    print("********************************************")

    # if the data file is missing, get it.
    if not os.path.exists("./GLODAPv2_pH_DIC_ALK_subset.csv"):
        print("Getting GLODAPv2 Data...")
        from get_GLODAP_data import get_GLODAP

        get_GLODAP()

    # Load data
    print("Importing GLODAPv2 Data...")
    gd = pd.read_csv("./GLODAPv2_pH_DIC_ALK_subset.csv")
    gd.dropna(
        subset=[
            "phtsinsitutp",
            "temperature",
            "salinity",
            "tco2",
            "talk",
            "pressure",
            "phosphate",
            "silicate",
        ],
        inplace=True,
    )
    gd.pressure /= 10  # convert pressure to bar

    # exclude weird cruise 270 data
    gd = gd.loc[gd.cruise != 270]

    # Do the work...
    print("Calculating pH from DIC and TA...")
    cpH = cb.Csys(
        TA=gd.talk,
        DIC=gd.tco2,
        T_in=gd.temperature,
        S_in=gd.salinity,
        P_in=gd.pressure,
        PT=gd.phosphate,
        SiT=gd.silicate,
        BT=415.7,
    )
    print("   Making plots...")
    fig, axs = cplot(gd.phtsinsitutp, cpH.pHtot, "pH", "Depth", gd.depth)
    fig.savefig(figdir + "/Figures/pH_comparison.png", dpi=200)

    print("Calculating TA from pH and DIC...")
    cTA = cb.Csys(
        pHtot=gd.phtsinsitutp,
        DIC=gd.tco2,
        T_in=gd.temperature,
        S_in=gd.salinity,
        P_in=gd.pressure,
        PT=gd.phosphate,
        SiT=gd.silicate,
        BT=415.7,
    )
    print("   Making plots...")
    fig, ax = cplot(gd.talk, cTA.TA, "Alk", "Depth", gd.depth)
    fig.savefig("Figures/TA_comparison.png", dpi=200)

    print("Calculating DIC from pH and TA...")
    cDIC = cb.Csys(
        pHtot=gd.phtsinsitutp,
        TA=gd.talk,
        T_in=gd.temperature,
        S_in=gd.salinity,
        P_in=gd.pressure,
        PT=gd.phosphate,
        SiT=gd.silicate,
        BT=415.7,
    )
    print("   Making plots...")
    fig, ax = cplot(gd.tco2, cDIC.DIC, "DIC", "Depth", gd.depth)
    fig.savefig("Figures/DIC_comparison.png", dpi=200)

    print("Done.")
    print()
    print(" > Plots are saved in : " + figdir + "/Figures/")
    print("********************************************")


# plotting function
def cplot(obs, pred, var, cvar, c, alpha=0.4, pclims=[0.05, 99.95]):
    fig = plt.figure(figsize=[11, 3.5])

    ax1 = fig.add_axes([0.07, 0.13, 0.33, 0.8])
    ax2 = fig.add_axes([0.47, 0.13, 0.33, 0.8])
    hax = fig.add_axes([0.805, 0.13, 0.08, 0.8])
    cax = fig.add_axes([0.9, 0.13, 0.015, 0.8])

    ad = np.concatenate([obs, pred])
    fna = np.isnan(ad)
    mn, mx = np.percentile(ad[~fna], pclims)
    xr = mx - mn
    pad = 0.1 * xr

    # Measured vs Predicted
    cm = ax1.scatter(obs, pred, s=5, alpha=alpha, c=c, lw=0)
    ax1.set_xlim(mn - pad, mx + pad)
    ax1.set_ylim(mn - pad, mx + pad)
    ax1.plot(
        ax1.get_xlim(), ax1.get_xlim(), ls="dashed", c=(0.5, 0.5, 0.5, 0.8), zorder=1
    )

    ax1.set_xlabel("GLODAPv2 Measured")
    ax1.set_ylabel("cbsyst predicted")

    ax1.text(
        0.05,
        0.95,
        "GLODAPv2 " + var,
        transform=ax1.transAxes,
        ha="left",
        va="top",
        fontsize=14,
        weight="bold",
        color=(0.4, 0.4, 0.4),
    )

    # Measured vs Difference
    diff = obs - pred
    fna = np.isnan(diff)
    mn, mx = np.percentile(diff[~fna], pclims)
    xr = mx - mn
    pad = 0.15 * xr

    # Scatterplot
    ax2.scatter(obs, diff, s=5, alpha=alpha, c=c, lw=0)
    ax2.set_xlim(ax1.get_xlim())
    ax2.set_ylim(mn - pad, mx + pad)
    ax2.axhline(0, ls="dashed", c=(0.5, 0.5, 0.5, 0.8), zorder=1)

    ax2.set_xlabel("GLODAPv2 Measured")
    ax2.set_ylabel("measured - predicted")

    # Histogram
    bins = np.linspace(*ax2.get_ylim(), 100)
    hax.hist(diff, bins=bins, orientation="horizontal", color=(0.6, 0.6, 0.6))
    hax.set_ylim(ax2.get_ylim())
    hax.set_xlabel("n")
    hax.set_yticklabels([])

    # Stats
    median = np.median(diff[~fna])
    pc95 = np.percentile(diff[~fna], [2.5, 97.5])
    hax.axhline(median, color="r", ls="dashed", zorder=2)
    hax.axhspan(*pc95, color="r", alpha=0.2, zorder=1)
    hax.set_ylim(ax2.get_ylim())
    hax.axhline(0, ls="dashed", c=(0.5, 0.5, 0.5, 0.8), zorder=1)

    if abs(median) < 0.01:
        number_fmt = '.1e'
    else:
        number_fmt = '.2f'

    ax2.text(
        0.03,
        0.97,
        f"Median Offset: {median:{number_fmt}}"
        + f"\n95% Limits: {pc95[0] - median:{number_fmt}} / +{pc95[1] - median:{number_fmt}}".format(*()),
        transform=ax2.transAxes,
        va="top",
        ha="left",
        backgroundcolor=(1, 1, 1, 0.5),
    )

    if not isinstance(c, str):
        fig.colorbar(cm, cax=cax, label=cvar)
    else:
        cax.set_visible(False)

    return fig, (ax1, ax2, hax, cax)

# plotting function
def cplot_resid(cb_pred, cb_var, glodap_data, glodap_var, cb_cvar, alpha=0.4, pclims=[0.05, 99.95]):
    fig = plt.figure(figsize=[11, 7])

    ax1 = fig.add_axes([0.07, 0.53, 0.33, 0.4])
    ax2 = fig.add_axes([0.47, 0.53, 0.33, 0.4])
    hax = fig.add_axes([0.805, 0.53, 0.08, 0.4])
    cax = fig.add_axes([0.9, 0.53, 0.015, 0.4])
    
    rax1 = fig.add_axes([0.07, 0.13, 0.25, 0.3])
    rax2 = fig.add_axes([0.37, 0.13, 0.25, 0.3])
    rax3 = fig.add_axes([0.67, 0.13, 0.25, 0.3])

    obs = glodap_data[glodap_var]
    pred = cb_pred[cb_var]
    c = cb_pred[cb_cvar]
    
    ad = np.concatenate([obs, pred])
    fna = np.isnan(ad)
    mn, mx = np.percentile(ad[~fna], pclims)
    xr = mx - mn
    pad = 0.1 * xr

    # Measured vs Predicted
    cm = ax1.scatter(obs, pred, s=5, alpha=alpha, c=c, lw=0)
    ax1.set_xlim(mn - pad, mx + pad)
    ax1.set_ylim(mn - pad, mx + pad)
    ax1.plot(
        ax1.get_xlim(), ax1.get_xlim(), ls="dashed", c=(0.5, 0.5, 0.5, 0.8), zorder=1
    )

    ax1.set_xlabel("GLODAPv2 Measured")
    ax1.set_ylabel("cbsyst predicted")

    ax1.text(
        0.05,
        0.95,
        "GLODAPv2 " + cb_var,
        transform=ax1.transAxes,
        ha="left",
        va="top",
        fontsize=14,
        weight="bold",
        color=(0.4, 0.4, 0.4),
    )

    # Measured vs Difference
    diff = obs - pred
    fna = np.isnan(diff)
    mn, mx = np.percentile(diff[~fna], pclims)
    xr = mx - mn
    pad = 0.15 * xr

    # Scatterplot
    ax2.scatter(obs, diff, s=5, alpha=alpha, c=c, lw=0)
    ax2.set_xlim(ax1.get_xlim())
    ax2.set_ylim(mn - pad, mx + pad)
    ax2.axhline(0, ls="dashed", c=(0.5, 0.5, 0.5, 0.8), zorder=1)

    ax2.set_xlabel("GLODAPv2 Measured")
    ax2.set_ylabel("measured - predicted")

    # Histogram
    bins = np.linspace(*ax2.get_ylim(), 100)
    hax.hist(diff, bins=bins, orientation="horizontal", color=(0.6, 0.6, 0.6))
    hax.set_ylim(ax2.get_ylim())
    hax.set_xlabel("n")
    hax.set_yticklabels([])

    # Stats
    median = np.median(diff[~fna])
    pc95 = np.percentile(diff[~fna], [2.5, 97.5])
    hax.axhline(median, color="r", ls="dashed", zorder=2)
    hax.axhspan(*pc95, color="r", alpha=0.2, zorder=1)
    hax.set_ylim(ax2.get_ylim())
    hax.axhline(0, ls="dashed", c=(0.5, 0.5, 0.5, 0.8), zorder=1)

    if abs(median) < 0.01:
        number_fmt = '.1e'
    else:
        number_fmt = '.2f'
    
    ax2.text(
        0.03,
        0.97,
        f"Median Offset: {median:{number_fmt}}"
        + f"\n95% Limits: {pc95[0] - median:{number_fmt}} / +{pc95[1] - median:{number_fmt}}".format(*()),
        transform=ax2.transAxes,
        va="top",
        ha="left",
        backgroundcolor=(1, 1, 1, 0.5),
    )

    if not isinstance(c, str):
        fig.colorbar(cm, cax=cax, label=cb_cvar)
    else:
        cax.set_visible(False)

    ymn, ymx = np.percentile(diff[~fna], pclims)
    yr = ymx - ymn
    ypad = 0.15 * yr
    for var, ax in zip(['T_in', 'S_in', 'P_in'], [rax1, rax2, rax3]):
        ax.scatter(cb_pred[var], diff, s=5, alpha=alpha, c=c, lw=0)

        mn, mx = np.percentile(cb_pred[var][~fna], pclims)                
        ax.set_xlim(mn, mx)
        ax.set_ylim(ymn - ypad, ymx + ypad)
        
        ax.set_xlabel(var)
        hax.axhline(0, ls="dashed", c=(0.5, 0.5, 0.5, 0.8), zorder=1)

    rax1.set_ylabel("measured - predicted")

    return fig, (ax1, ax2, hax, cax), (rax1, rax2, rax3)


if __name__ == "__main__":
    GLODAPv2_comparison()



import os
import zipfile
import urllib.request as ureq

import pandas as pd
import numpy as np

from tqdm import tqdm


# Progress Bar
class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


def get_GLODAP(path="./", leave_zip=True):
    if not os.path.exists(path + "/GLODAPv2 Merged Master File.csv.zip"):
        print("Fetching GLODAPv2 Data...")

        GLODAP_urls = [
            "https://www.glodap.info/glodap_files/v2.2020/GLODAPv2.2020_Merged_Master_File.csv.zip",
            # "https://www.glodap.info/glodap_files/v2.2023/GLODAPv2.2023_Merged_Master_File.csv.zip",
        ]
        downloaded = False
        # download GLODAP data
        i = 0
        while not downloaded:
            GLODAP_url = GLODAP_urls[i]
            try:
                with TqdmUpTo(
                    unit="B", unit_scale=True, miniters=1, desc="Downloading GLODAPv2"
                ) as t:
                    ureq.urlretrieve(
                        GLODAP_url,
                        path + "/GLODAPv2 Merged Master File.csv.zip",
                        reporthook=t.update_to,
                    )
                downloaded = True
            except:
                pass
            i += 1

        # # open URL
        # file = requests.get(GLODAP_url, stream=True)
        # total_size = int(file.headers.get('content-length', 0))

        # # Download data
        # with open('./GLODAPv2 Merged Master File.csv.zip', 'wb') as f:
        #     for data in tqdm(file.iter_content(1024), total=total_size / (1024), unit='KB', unit_scale=True):
        #         f.write(data)
    else:
        print("Found GLODAPv2 Data...")

    print("Reading data...")
    # open zip
    zf = zipfile.ZipFile(path + "/GLODAPv2 Merged Master File.csv.zip")

    # read data into pandas
    gd = pd.read_csv(zf.open("GLODAPv2.2020_Merged_Master_File.csv"))
    # gd = pd.read_csv(zf.open("GLODAPv2.2023_Merged_Master_File.csv"))

    # replace missing values with nan
    gd.replace(-9999, np.nan, inplace=True)

    print("Selecting 'good' (flag == 2) data...")
    # isolate good data only (flag = 2)
    # gd.loc[gd.phts25p0f != 2, 'phts25p0'] = np.nan
    gd.loc[gd.phtsinsitutpf != 2, "phtsinsitutp"] = np.nan
    gd.loc[gd.tco2f != 2, "tco2"] = np.nan
    gd.loc[gd.talkf != 2, "talk"] = np.nan
    gd.loc[gd.salinityf != 2, "salinity"] = np.nan
    gd.loc[gd.phosphatef != 2, "phosphate"] = np.nan
    gd.loc[gd.silicatef != 2, "silicate"] = np.nan

    # Identify rows where ph, dic, talk and sal are present
    # phind = ~gd.phtsinsitutp.isnull()
    # dicind = ~gd.tco2.isnull()
    # alkind = ~gd.talk.isnull()
    # salind = ~gd.salinity.isnull()

    gd.dropna(
        subset=[
            "phtsinsitutp",
            "tco2",
            "talk",
            "temperature",
            "salinity",
            "pressure",
            "silicate",
            "phosphate",
        ],
        inplace=True,
    )

    print("Saving data subset...")
    # Isolate those data
    # gds = gd.loc[phind & dicind & alkind & salind, ['phts25p0', 'phtsinsitutp', 'tco2', 'talk', 'temperature', 'salinity',
    #                                                 'cruise', 'station', 'cast', 'year', 'month', 'day', 'hour',
    #                                                 'latitude', 'longitude', 'bottomdepth', 'maxsampdepth', 'bottle',
    #                                                 'pressure', 'depth', 'theta', 'silicate', 'phosphate']]
    gds = gd.loc[
        :,
        [
            "phts25p0",
            "phtsinsitutp",
            "tco2",
            "talk",
            "temperature",
            "salinity",
            "cruise",
            "station",
            "cast",
            "year",
            "month",
            "day",
            "hour",
            "latitude",
            "longitude",
            "bottomdepth",
            "maxsampdepth",
            "bottle",
            "pressure",
            "depth",
            "theta",
            "silicate",
            "phosphate",
        ],
    ]

    gds.to_csv(path + "/GLODAPv2_pH_DIC_ALK_subset.csv", index=False)

    if not leave_zip:
        os.remove(path + "/GLODAPv2 Merged Master File.csv.zip")

    return


if __name__ == "__main__":
    print("\n********************************************")
    print("Get GLODAPv2 carbon data for cbsyst testing.")
    print("********************************************")
    get_GLODAP()
    print("Done.")
    print("********************************************\n")



import unittest
from cbsyst import carbon as cf

from tests import check_vals
ref = check_vals.carbon_ref

class CarbonFnTestCase(unittest.TestCase):
    """Test all C functions"""

    def test_Carbon(self):
        
        with self.subTest(msg='CO2_pH'):
            self.assertAlmostEqual(
                cf.CO2_pH(CO2=ref.CO2, pH=ref.pHtot, Ks=ref.Ks), ref.DIC, places=6
            )

        with self.subTest(msg='CO2_HCO3 (zf)'):
            self.assertAlmostEqual(
                cf.CO2_HCO3(CO2=ref.CO2, HCO3=ref.HCO3, Ks=ref.Ks), ref.H, places=6
            )

        with self.subTest(msg='CO2_CO3 (zf)'):
            self.assertAlmostEqual(
                cf.CO2_CO3(CO2=ref.CO2, CO3=ref.CO3, Ks=ref.Ks), ref.H, places=6
            )

        with self.subTest(msg='CO2_TA'):
            self.assertAlmostEqual(
                cf.CO2_TA(
                    CO2=ref.CO2 / ref.unit,
                    TA=ref.TA / ref.unit,
                    BT=ref.BT / ref.unit,
                    PT=ref.TP / ref.unit,
                    SiT=ref.TSi / ref.unit,
                    ST=ref.TS,
                    FT=ref.TF,
                    Ks=ref.Ks,
                )[0],
                ref.pHtot,
                places=6,
            )
        
        with self.subTest(msg='CO2_DIC (zf)'):        
            self.assertAlmostEqual(
                cf.CO2_DIC(CO2=ref.CO2, DIC=ref.DIC, Ks=ref.Ks), ref.H, places=6
            )

        with self.subTest(msg='pH_HCO3'):        
            self.assertAlmostEqual(
                cf.pH_HCO3(pH=ref.pHtot, HCO3=ref.HCO3, Ks=ref.Ks), ref.DIC, places=6
            )

        with self.subTest(msg='pH_CO3'):        
            self.assertAlmostEqual(
                cf.pH_CO3(pH=ref.pHtot, CO3=ref.CO3, Ks=ref.Ks), ref.DIC, places=6
            )

        with self.subTest(msg='pH_TA'):        
            self.assertAlmostEqual(
                cf.pH_TA(
                    pH=ref.pHtot,
                    TA=ref.TA / ref.unit,
                    BT=ref.BT / ref.unit,
                    PT=ref.TP / ref.unit,
                    SiT=ref.TSi / ref.unit,
                    ST=ref.TS,
                    FT=ref.TF,
                    Ks=ref.Ks,
                )
                * ref.unit,
                ref.DIC,
                places=6,
            )

        with self.subTest(msg='pH_DIC'):        
            self.assertAlmostEqual(
                cf.pH_DIC(pH=ref.pHtot, DIC=ref.DIC, Ks=ref.Ks), ref.CO2, places=6
            )

        with self.subTest(msg='HCO3_CO3 (zf)'):        
            self.assertAlmostEqual(
                cf.HCO3_CO3(HCO3=ref.HCO3, CO3=ref.CO3, Ks=ref.Ks), ref.H, places=6
            )

        with self.subTest(msg='HCO3_TA (zf)'):        
            self.assertAlmostEqual(
                cf.HCO3_TA(HCO3=ref.HCO3 / ref.unit, TA=ref.TA / ref.unit, BT=ref.BT / ref.unit, Ks=ref.Ks),
                ref.H,
                places=6,
            )

        with self.subTest(msg='HCO3_DIC (zf)'):        
            self.assertAlmostEqual(
                cf.HCO3_DIC(HCO3=ref.HCO3, DIC=ref.DIC, Ks=ref.Ks), ref.H, places=6
            )

        with self.subTest(msg='CO3_TA (zf)'):        
            self.assertAlmostEqual(
                cf.CO3_TA(CO3=ref.CO3 / ref.unit, TA=ref.TA / ref.unit, BT=ref.BT / ref.unit, Ks=ref.Ks),
                ref.H,
                places=6,
            )

        with self.subTest(msg='CO3_DIC (zf)'):        
            self.assertAlmostEqual(
                cf.CO3_DIC(CO3=ref.CO3, DIC=ref.DIC, Ks=ref.Ks), ref.H, places=6
            )

        with self.subTest(msg='TA_DIC'):        
            self.assertAlmostEqual(
                cf.TA_DIC(
                    TA=ref.TA / ref.unit,
                    DIC=ref.DIC / ref.unit,
                    BT=ref.BT / ref.unit,
                    PT=ref.TP / ref.unit,
                    SiT=ref.TSi / ref.unit,
                    ST=ref.TS,
                    FT=ref.TF,
                    Ks=ref.Ks,
                )[0],
                ref.pHtot,
                places=6,
            )

        with self.subTest(msg='cCO2'):        
            self.assertAlmostEqual(
                cf.cCO2(H=ref.H, DIC=ref.DIC, Ks=ref.Ks), ref.CO2, places=6
            )

        with self.subTest(msg='cCO3'):        
            self.assertAlmostEqual(
                cf.cCO3(H=ref.H, DIC=ref.DIC, Ks=ref.Ks), ref.CO3, places=6
            )

        with self.subTest(msg='cHCO3'):        
            self.assertAlmostEqual(
                cf.cHCO3(H=ref.H, DIC=ref.DIC, Ks=ref.Ks), ref.HCO3, places=6
            )

        with self.subTest(msg='cTA'):        
            (TA, CAlk, BAlk, PAlk, SiAlk, OH, Hfree, HSO4, HF) = cf.cTA(
                H=ref.H,
                DIC=ref.DIC / ref.unit,
                BT=ref.BT / ref.unit,
                PT=ref.TP / ref.unit,
                SiT=ref.TSi / ref.unit,
                ST=ref.TS,
                FT=ref.TF,
                Ks=ref.Ks,
                mode="multi",
            )

            self.assertAlmostEqual(TA * ref.unit, ref.TA, msg="cTA - TA", places=6)
            self.assertAlmostEqual(CAlk * ref.unit, ref.CAlk, msg="cTA - CAlk", places=6)
            self.assertAlmostEqual(BAlk * ref.unit, ref.BAlk, msg="cTA - BAlk", places=6)
            self.assertAlmostEqual(PAlk * ref.unit, ref.PAlk, msg="cTA - PAlk", places=6)
            self.assertAlmostEqual(SiAlk * ref.unit, ref.SiAlk, msg="cTA - SiAlk", places=6)
            self.assertAlmostEqual(OH * ref.unit, ref.OH, msg="cTA - OH", places=6)
            self.assertAlmostEqual(Hfree * ref.unit, ref.Hfree, msg="cTA - Hfree", places=6)
            self.assertAlmostEqual(HSO4 * ref.unit, ref.HSO4, msg="cTA - HSO4", places=6)
            self.assertAlmostEqual(HF * ref.unit, ref.HF, msg="cTA - HF", places=6)

        with self.subTest(msg='fCO2_to_CO2'):        
            self.assertAlmostEqual(
                cf.fCO2_to_CO2(fCO2=ref.fCO2, Ks=ref.Ks), ref.CO2, places=6
            )

        with self.subTest(msg='CO2_to_fCO2'):        
            self.assertAlmostEqual(
                cf.CO2_to_fCO2(CO2=ref.CO2, Ks=ref.Ks), ref.fCO2, places=6
            )

        with self.subTest(msg='fCO2_to_pCO2'):        
            self.assertAlmostEqual(
                cf.fCO2_to_pCO2(fCO2=ref.fCO2, Tc=ref.T), ref.pCO2, places=6
            )

        with self.subTest(msg='pCO2_to_fCO2'):        
            self.assertAlmostEqual(
                cf.pCO2_to_fCO2(pCO2=ref.pCO2, Tc=ref.T), ref.fCO2, places=6
            )




"""
TODO: Script to compare CO2SYS.m output to cbsyst output.
Comparison similar to Orr et al (2015)

Parameters
----------
raw_data : str
    Path to GLODAP-WOA comparison data.
co2sys_data : str
    Path to output from CO2SYS.m
output_dir : str
    Output directory for plots

Returns
-------
Prints a range of statistics comparing CO2SYS and cbsyst.
Creates a variety of plots with statistics in output_dir.
"""



import uncertainties.unumpy as unp
import kgen
import numpy as np
import pandas as pd


# Helpers useful to the user
# --------------------------
def data_out(cbdat, path=None, include_constants=False):
    """
    Save output from cbsyst.

    Parameters
    ----------
    cbdat : dict / Bunch
        The output from Csys, Bsys, ABsys or CBsys.
    path : str
        The file name (and path) where you want to
        save the data. If not provided, data are not
        saved to a file.

        The extension of the file determines the output
        format. Can be 'csv', 'xls', 'html, 'tex', or 'pkl'.
    include_constants : bool
        If True, include pK and alpha constants in output.

    Returns
    -------
    * pandas.DataFrame of output
    * Saves file (if specified)

    """

    cols = [
        "pHtot",
        "pHfree",
        "pHsws",
        "pHNBS",
        "DIC",
        "fCO2",
        "pCO2",
        "CO2",
        "HCO3",
        "CO3",
        "TA",
        "BT",
        "BO3",
        "BO4",
        "dBT",
        "dBO3",
        "dBO4",
        "ABT",
        "ABO3",
        "ABO4",
        "T_in",
        "S_in",
        "P_in",
        "T_out",
        "S_out",
        "P_out",
        "Ca",
        "Mg",
    ]

    consts = ["K0", "K1", "K2", "KB", "KW", "KS", "KspA", "KspC"]

    size = cbdat.pHtot.size
    out = pd.DataFrame(index=range(size))

    for c in cols:
        if c in cbdat and cbdat[c] is not None:
            if (np.ndim(cbdat[c]) == 1) and (cbdat[c].size == 1):
                cbdat[c] = cbdat[c][0]
            if c in cbdat:
                out.loc[:, c] = cbdat[c]

    if include_constants:
        for c in consts:
            if c in cbdat.Ks and cbdat.Ks[c] is not None:
                if (np.ndim(cbdat.Ks[c]) == 1) and (cbdat.Ks[c].size == 1):
                    cbdat.Ks[c] = cbdat.Ks[c][0]
                out.loc[:, "p" + c] = -np.log10(cbdat.Ks[c])
        if "alphaB" in cbdat and cbdat.alphaB is not None:
            if (np.ndim(cbdat.alphaB) == 1) and (cbdat.alphaB.size == 1):
                cbdat.alphaB = cbdat.alphaB[0]
            out.loc[:, "alphaB"] = cbdat.alphaB

    if path is not None:
        fmt = path.split(".")[-1]
        fdict = {
            "csv": "to_csv",
            "html": "to_html",
            "xls": "to_excel",
            "pkl": "to_pickle",
            "tex": "to_latex",
        }

        if fmt not in fdict:
            raise ValueError(
                (
                    "File extension does not match available output\n"
                    + "options. Should be one of 'csv', 'html', 'xls',\n"
                    + "'pkl' (pickle) or 'tex' (LaTeX)."
                )
            )
        try:
            _ = getattr(out, fdict[fmt])(path, index=None)
        except TypeError:
            _ = getattr(out, fdict[fmt])(path)

    return out


# Programmatic helpers for code elsewhere
# ---------------------------------------

# Bunch modifies dict to allow item access using dot (.) operator
class Bunch(dict):
    def __init__(self, *args, **kwds):
        super(Bunch, self).__init__(*args, **kwds)
        self.__dict__ = self


def noms(*it):
    """
    Return nominal_values for provided objects.

    Parameters
    ----------
    *it : n objects
    """
    return [unp.nominal_values(i) for i in it]


def maxL(*it):
    """
    Calculate maximum length of provided items.

    Parameters
    ----------
    *it : objects
        Items of various lengths. Only lengths
        of iterables are returned.

    Returns
    -------
    Length of longest object (int).
    """
    m = set()
    for i in it:
        try:
            m.add(len(i))
        except TypeError:
            pass
    if len(m) > 0:
        return max(m)
    else:
        return 1
    
def maxSize(*it):
    """
    Calculate maximum size of provided items.

    Parameters
    ----------
    *it : objects
        Items of various sizes. Only sizes
        of iterables are returned.

    Returns
    -------
    size of largest object (int).
    """
    m = set()
    for i in it:
        try:
            m.add(np.size(i))
        except TypeError:
            pass
    if np.size(m) > 0:
        return max(m)
    else:
        return 1
    
def maxD(*it):
    """
    Calculate maximum number of dimensions in provided items.
    
    Parameters
    ----------
    *it : objects
        Items of various shapes with an .ndim attribute.
    """
    return np.max([x.ndim for x in it])

def maxShape(*it):
    """
    Returns the shape of the largest array.
    """
    size = 0
    shape = None
    for i in it:
        i = np.asanyarray(i)
        if i.size > size:
            size = i.size
            shape = i.shape
    return shape

def cast_array(*it):
    """
    Recasts inputs into array of shape (len(it), maxSize(*it))
    """
    new = np.zeros((len(it), maxSize(*it)))
    for i, t in enumerate(it):
        new[i, :] = np.ravel(t)
    return new


def NnotNone(*it):
    """
    Returns the number of elements of it tha are not None.

    Parameters
    ----------
    it : iterable
        iterable of elements that are either None, or not None

    Returns
    -------
    int
    """
    return sum([i is not None for i in it])

# Helpers for aspects of seawater chemistry
# -----------------------------------------
def prescorr(P, Tc, a0, a1, a2, b0, b1):
    """
    Calculate pressore correction factor for thermodynamic Ks.

    From Millero et al (2007, doi:10.1021/cr0503557)
    Eqns 38-40

    Usage:
    K_corr / K_orig = [output]
    Kcorr = [output] * K_orig
    """
    dV = a0 + a1 * Tc + a2 * Tc ** 2
    dk = (b0 + b1 * Tc) / 1000
    # factor of 1000 not mentioned in Millero,
    # but present in Zeebe book, and used in CO2SYS
    RT = 83.1451 * (Tc + 273.15)
    return np.exp((-dV + 0.5 * dk * P) * P / RT)


def swdens(TempC, Sal):
    """
    Seawater Density (kg / L) from Temp (C) and Sal (PSU)

    Chapter 5, Section 4.2 of Dickson, Sabine and Christian
    (2007, http://cdiac.ornl.gov/oceans/Handbook_2007.html)

    Parameters
    ----------
    TempC : array-like
        Temperature in celcius.
    Sal : array-like
        Salinity in PSU

    Returns
    -------
    Density in kg / L
    """
    # convert temperature to IPTS-68
    T68 = (TempC + 0.0002) / 0.99975
    pSMOW = (
        999.842594
        + 6.793952e-2 * T68
        + -9.095290e-3 * T68 ** 2
        + 1.001685e-4 * T68 ** 3
        + -1.120083e-6 * T68 ** 4
        + 6.536332e-9 * T68 ** 5
    )
    A = (
        8.24493e-1
        + -4.0899e-3 * T68
        + 7.6438e-5 * T68 ** 2
        + -8.2467e-7 * T68 ** 3
        + 5.3875e-9 * T68 ** 4
    )
    B = -5.72466e-3 + 1.0227e-4 * T68 + -1.6546e-6 * T68 ** 2
    C = 4.8314e-4
    return (pSMOW + A * Sal + B * Sal ** 1.5 + C * Sal ** 2) / 1000


def calc_ST(Sal):
    """
    Calculate total Sulphur in mol/kg-SW- lifted directly from CO2SYS.m

    From Dickson et al., 2007, Table 2
    Note: Sal / 1.80655 = Chlorinity
    """
    return 0.14 * Sal / 1.80655 / 96.062 # mol/kg-SW


def calc_FT(Sal):
    """
    Calculate total Fluorine in mol/kg-SW

    From Dickson et al., 2007, Table 2
    Note: Sal / 1.80655 = Chlorinity
    """
    return 6.7e-5 * Sal / 1.80655 / 18.9984 # mol/kg-SW


# def calc_BT(Sal):
#     """
#     Calculate total Boron

#     Lee, Kim, Byrne, Millero, Feely, Yong-Ming Liu. 2010.
#     Geochimica Et Cosmochimica Acta 74 (6): 1801-1811
#     """
#     a, b = (0.0004326, 35.)
#     return a * Sal / b


def calc_BT(Sal):
    """
    Calculate total Boron in mol/kg-SW - lifted directly from CO2SYS.m

    Directly from CO2SYS:
    Uppstrom, L., Deep-Sea Research 21:161-162, 1974:
    this is 0.000416 * Sal/35. = 0.0000119 * Sal
    TB(FF) = (0.000232 / 10.811) * (Sal / 1.80655) in mol/kg-SW
    """
    a, b = (0.0004157, 35.0)
    return a * Sal / b  # mol/kg-SW


def calc_fH(TempK, Sal):
    # Same as CO2SYS
    # Takahashi et al, Chapter 3 in GEOSECS Pacific Expedition,
    # v. 3, 1982 (p. 80)

    a, b, c, d = (1.2948, -2.036e-3, 4.607e-4, -1.475e-6)
    return a + b * TempK + (c + d * TempK) * Sal ** 2


# Convert between pH scales
def calc_pH_scales(pHtot, pHfree, pHsws, pHNBS, ST, FT, TempK, Sal, Ks):
    """
    Calculate pH on all scales, given one.
    """

    # check if any pH scale is given.
    npH = NnotNone(pHfree, pHsws, pHtot, pHNBS)

    if npH == 1:
        # pH scale conversions
        FREEtoTOT = -np.log10((1 + ST / Ks.KS))
        SWStoTOT = -np.log10((1 + ST / Ks.KS) / (1 + ST / Ks.KS + FT / Ks.KF))
        fH = calc_fH(TempK, Sal)

        if pHtot is not None:
            return {
                "pHfree": pHtot - FREEtoTOT,
                "pHsws": pHtot - SWStoTOT,
                "pHNBS": pHtot - SWStoTOT - np.log10(fH),
            }
        elif pHsws is not None:
            return {
                "pHfree": pHsws + SWStoTOT - FREEtoTOT,
                "pHtot": pHsws + SWStoTOT,
                "pHNBS": pHsws - np.log10(fH),
            }
        elif pHfree is not None:
            return {
                "pHsws": pHfree + FREEtoTOT - SWStoTOT,
                "pHtot": pHfree + FREEtoTOT,
                "pHNBS": pHfree + FREEtoTOT - SWStoTOT - np.log10(fH),
            }
        elif pHNBS is not None:
            return {
                "pHsws": pHNBS + np.log10(fH),
                "pHtot": pHNBS + np.log10(fH) + SWStoTOT,
                "pHfree": pHNBS + np.log10(fH) + SWStoTOT - FREEtoTOT,
            }
    else:
        return {}

def pH_scale_converter(pH, scale, Temp, Sal, Press=None, ST=None, FT=None):
    """
    Returns pH on all scales.
    """
    pH_scales = ["Total", "FREE", "SWS", "NBS"]
    if scale not in pH_scales:
        raise ValueError("scale must be one of Total, NBS, SWS or FREE.")
    if ST is None:
        ST = calc_ST(Sal)
    if FT is None:
        FT = calc_FT(Sal)
    TempK = Temp + 273.15

    Ks = Bunch(kgen.calc_Ks(temp_c=Temp, sal=Sal, p_bar=Press))

    inp = [None, None, None, None]
    inp[np.argwhere(scale == np.array(pH_scales))[0, 0]] = pH

    return calc_pH_scales(*inp, ST, FT, TempK, Sal, Ks)


# # TODO: function that correct pH for temperature
# def calc_pH_Tcorr(pH, T_in, T_out, scale, Sal, Press=None, ST=None, FT=None):
#     """
#     Returns pH on all scales.
#     """
#     pH_scales = ["Total", "FREE", "SWS", "NBS"]
#     if scale not in pH_scales:
#         raise ValueError("scale must be one of Total, NBS, SWS or FREE.")
#     if ST is None:
#         ST = calc_ST(Sal)
#     if FT is None:
#         FT = calc_FT(Sal)

#     Ks_in = Bunch(kgen.calc_Ks(temp_c=T_in, sal=Sal, p_bar=Press))
#     Ks_out = Bunch(kgen.calc_Ks(temp_c=T_out, sal=Sal, p_bar=Press))

#     inp = [None, None, None, None]
#     inp[np.argwhere(scale == np.array(pH_scales))[0, 0]] = pH

#     pH_in = calc_pH_scales(*inp, ST, FT, T_in + 273.15, Sal, Ks_in)
#     pH_out = calc_pH_scales(*inp, ST, FT, T_out + 273.15, Sal, Ks_out)

#     return pH_in, pH_out